from .Graphable import Graphable
from .Printable import Printable
from .Network import Network
from .AddressAssignmentConstraint import AddressAssignmentConstraint
from .Registry import ScopedRegistry
from .enums import NetworkType, NodeRole
from .Node import Node
from ipaddress import IPv4Network
from typing import Generator, Dict


class AutonomousSystem(Printable, Graphable):
    """!
    @brief AutonomousSystem class. 

    This class represents an autonomous system.
    """

    __asn: int
    __subnet_generator: Generator[IPv4Network, None, None]
    __reg: ScopedRegistry

    def __init__(self, asn: int):
        """!
        @brief AutonomousSystem constructor.

        @param asn ASN for this system.
        """
        Graphable.__init__(self)
        self.__asn = asn
        self.__reg = ScopedRegistry(str(asn))
        self.__subnet_generator = None if asn > 255 else IPv4Network("10.{}.0.0/16".format(asn)).subnets(new_prefix = 24)

    def getAsn(self) -> int:
        """!
        @brief Get ASN.
        """
        return self.__asn
    
    def createNetwork(self, name: str, prefix: str = "auto", aac: AddressAssignmentConstraint = None) -> Network:
        """!
        @brief Create a new network.

        @param name name of the new network.
        @param prefix optional. Network prefix of this network. If not set, a
        /24 subnet of "10.{asn}.{id}.0/24" will be used, where asn is ASN of
        this AS, and id is a self-incremental value starts from 0.
        @param aac optional. AddressAssignmentConstraint to use.
        @throws StopIteration if subnet exhausted.
        """
        assert prefix != "auto" or self.__asn <= 255, "can't use auto: asn > 255"

        network = IPv4Network(prefix) if prefix != "auto" else next(self.__subnet_generator)
        return self.__reg.register('net', name, Network(name, NetworkType.Local, network, aac))

    def getNetwork(self, name: str) -> Network:
        """!
        @brief Retrive a network.

        @param name name of the network.
        @returns Network.
        """
        return self.__reg.get('net', name)

    def createRouter(self, name: str) -> Node:
        """!
        @brief Create a router node.

        @param name name of the new node.
        @returns Node.
        """
        return self.__reg.register('rnode', name, Node(name, NodeRole.Router, self.__asn))

    def getRouter(self, name: str) -> Node:
        """!
        @brief Retrive a router node.

        @param name name of the node.
        @returns Node.
        """
        return self.__reg.get('rnode', name)

    def createHost(self, name: str) -> Node:
        """!
        @brief Create a host node.

        @param name name of the new node.
        @returns Node.
        """
        return self.__reg.register('hnode', name, Node(name, NodeRole.Host, self.__asn))

    def getHost(self, name: str) -> Node:
        """!
        @brief Retrive a host node.

        @param name name of the node.
        @returns Node.
        """
        return self.__reg.get('hnode', name)

    def _doCreateGraphs(self):
        l2graph = self._addGraph('AS{}: Layer 2 Connections'.format(self.__asn), False)
        
        for obj in self.__reg.getByType('net'):
            net: Network = obj
            l2graph.addVertex('Network: {}'.format(net.getName()), shape = 'rectangle', group = 'AS{}'.format(self.__asn))

        for obj in self.__reg.getByType('rnode'):
            router: Node = obj
            rtrname = 'Router: {}'.format(router.getName(), group = 'AS{}'.format(self.__asn))
            l2graph.addVertex(rtrname, group = 'AS{}'.format(self.__asn), shape = 'diamond')
            for iface in router.getInterfaces():
                net = iface.getNet()
                netname = 'Network: {}'.format(net.getName())
                if net.getType() == NetworkType.InternetExchange:
                    netname = 'Exchange: {}...'.format(net.getName())
                    l2graph.addVertex(netname, shape = 'rectangle')
                l2graph.addEdge(rtrname, netname)

        for obj in self.__reg.getByType('hnode'):
            router: Node = obj
            rtrname = 'Host: {}'.format(router.getName(), group = 'AS{}'.format(self.__asn))
            l2graph.addVertex(rtrname, group = 'AS{}'.format(self.__asn))
            for iface in router.getInterfaces():
                net = iface.getNet()
                netname = 'Network: {}'.format(net.getName())
                l2graph.addEdge(rtrname, netname)

        
    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'AutonomousSystem {}:\n'.format(self.__asn)

        indent += 4
        out += ' ' * indent
        out += 'Networks:\n'

        for net in self.__reg.getByType('net'):
            out += net.print(indent + 4)

        out += ' ' * indent
        out += 'Routers:\n'

        for net in self.__reg.getByType('rnode'):
            out += net.print(indent + 4)

        out += ' ' * indent
        out += 'Hosts:\n'

        for net in self.__reg.getByType('hnode'):
            out += net.print(indent + 4)

        return out


from typing import List, Dict, Tuple
from seedsim.core import Printable, Registrable
from sys import stderr
from enum import Enum


class Layer(Printable, Registrable):
    """!
    @brief The layer interface.
    """

    dependencies: Dict[str, List[Tuple[str, bool]]] = {}

    def addDependency(self, layerName: str, reverse: bool, optional: bool):
        """!
        @brief add layer dependency.

        @param layerName name of the layer.
        @param reverse add as reverse dependency. Regular dependency requires
        the given layer to be rendered before the current layer. Reverse
        dependency requires the given layer to be rendered after the current
        layer. 
        @param optional continue render even if the given layer does not exist.
        Does not work for reverse dependencies.
        """

        _current = layerName if reverse else self.getName()
        _target = self.getName() if reverse else layerName

        if _current not in self.dependencies:
            self.dependencies[_current] = []

        self.dependencies[_current].append((_target, optional))

    def getName(self) -> str:
        """!
        @brief Get name of this layer.

        This method should return a unique name for this layer. This will be
        used by the renderer to resolve dependencies relationships.

        @returns name of the layer.
        """
        raise NotImplementedError('getName not implemented')

    def onRender(self) -> None:
        """!
        @brief Handle rendering.
        """
        raise NotImplementedError('onRender not implemented')

    def _log(self, message: str) -> None:
        """!
        @brief Log to stderr.
        """
        print("==== {}Layer: {}".format(self.getName(), message), file=stderr)



from .Layer import Layer
from seedsim.core import Registry, ScopedRegistry, Node, Interface, Network
from seedsim.core.enums import NetworkType
from typing import List, Dict, Set
from functools import partial
from ipaddress import IPv4Network

RoutingFileTemplates: Dict[str, str] = {}

RoutingFileTemplates["protocol"] = """\
protocol {protocol} {name} {{{body}}}
"""

RoutingFileTemplates["pipe"] = """\
protocol pipe {{
    table {src};
    peer table {dst};
    import {importFilter};
    export {exportFilter};
}}
"""

RoutingFileTemplates["rs_bird"] = """\
router id {routerId};
protocol device {{
}}
"""

RoutingFileTemplates["rnode_bird"] = """\
router id {routerId};
ipv4 table t_direct;
protocol device {{
}}
protocol kernel {{
    ipv4 {{
        import all;
        export all;
    }};
    learn;
}}
"""

RoutingFileTemplates['rnode_bird_direct'] = """
    ipv4 {{
        table t_direct;
        import all;
    }};
{interfaces}
"""

RoutingFileTemplates["rnode_bird_direct_interface"] = """
    interface "{interfaceName}";
"""

class Router(Node):
    """!
    @brief Node extension class.

    Nodes with routing install will be replaced with this to get the extension
    methods.
    """

    __loopback_address: str

    def setLoopbackAddress(self, address: str):
        """!
        @brief Set loopback address.

        @param address address.
        """
        self.__loopback_address = address

    def getLoopbackAddress(self) -> str:
        """!
        @brief Get loopback address.

        @returns address.
        """
        return self.__loopback_address

    def addProtocol(self, protocol: str, name: str, body: str):
        """!
        @brief Add a new protocol to BIRD on the given node.

        @param protocol protocol type. (e.g., bgp, ospf)
        @param name protocol name.
        @param body protocol body.
        """
        self.appendFile("/etc/bird/bird.conf", RoutingFileTemplates["protocol"].format(
            protocol = protocol,
            name = name,
            body = body
        ))

    def addTablePipe(self, src: str, dst: str = 'master4', importFilter: str = 'none', exportFilter: str = 'all', ignoreExist: bool = True):
        """!
        @brief add a new routing table pipe.
        
        @param src src table.
        @param dst (optional) dst table (default: master4)
        @param importFilter (optional) filter for importing from dst table to src table (default: none)
        @param exportFilter (optional) filter for exporting from src table to dst table (default: all)
        @param ignoreExist (optional) assert check if table exists. If true, error is silently discarded.

        @throws AssertionError if pipe between two tables already exist and ignoreExist is False.
        """
        meta = self.getAttribute('__routing_layer_metadata', {})
        if 'pipes' not in meta: meta['pipes'] = {}
        pipes = meta['pipes']
        if src not in pipes: pipes[src] = []
        if dst in pipes[src]:
            assert ignoreExist, 'pipe from {} to {} already exist'.format(src, dst)
            return
        pipes[src].append(dst)
        self.appendFile('/etc/bird/bird.conf', RoutingFileTemplates["pipe"].format(
            src = src,
            dst = dst,
            importFilter = importFilter,
            exportFilter = exportFilter
        ))

    def addTable(self, tableName: str):
        """!
        @brief Add a new routing table to BIRD on the given node.

        @param tableName name of the new table.
        """
        meta = self.getAttribute('__routing_layer_metadata', {})
        if 'tables' not in meta: meta['tables'] = []
        tables = meta['tables']
        if tableName not in tables: self.appendFile('/etc/bird/bird.conf', 'ipv4 table {};\n'.format(tableName))
        tables.append(tableName)

class Routing(Layer):
    """!
    @brief The Routing layer.

    This layer provides routing support for routers and hosts. i.e., (1) install
    BIRD on router nodes and allow BGP/OSPF to work, (2) setup kernel and device
    protocols, and (3) setup defult routes for host nodes.

    When this layer is rendered, two new methods will be added to the router
    node and can be used by other layers: (1) addProtocol: add new protocol
    block to BIRD, and (2) addTable: add new routing table to BIRD.
    
    This layer also assign loopback address for iBGP/LDP, etc., for other
    protocols to use later and as router id.
    """

    __reg = Registry()
    __direct_nets: Set[Network]
    __loopback_assigner = IPv4Network('10.0.0.0/16')
    __loopback_pos = 1

    def __init__(self):
        """!
        @brief Routing layre constructor.
        """
        self.__direct_nets = set()
        self.addDependency('Base', False, False)
    
    def getName(self) -> str:
        return "Routing"

    def __installBird(self, node: Node):
        """!
        @brief Install bird on node, and handle the bug.
        """
        node.addBuildCommand('mkdir -p /usr/share/doc/bird2/examples/')
        node.addBuildCommand('touch /usr/share/doc/bird2/examples/bird.conf')
        node.addBuildCommand('apt-get install -y --no-install-recommends bird2')

    def onRender(self):
        for ((scope, type, name), obj) in self.__reg.getAll().items():
            if type == 'rs':
                rs_node: Node = obj
                self.__installBird(rs_node)
                rs_node.addStartCommand('[ ! -d /run/bird ] && mkdir /run/bird')
                rs_node.addStartCommand('bird -d', True)
                self._log("Bootstraping bird.conf for RS {}...".format(name))

                rs_ifaces = rs_node.getInterfaces()
                assert len(rs_ifaces) == 1, "rs node {} has != 1 interfaces".format(rs_node.getName())

                rs_iface = rs_ifaces[0]

                if not issubclass(rs_node.__class__, Router): rs_node.__class__ = Router
                rs_node.setFile("/etc/bird/bird.conf", RoutingFileTemplates["rs_bird"].format(
                    routerId = rs_iface.getAddress()
                ))
                
            if type == 'rnode':
                rnode: Router = obj
                if not issubclass(rnode.__class__, Router): rnode.__class__ = Router

                self._log("Setting up loopback interface for AS{} Router {}...".format(scope, name))

                lbaddr = self.__loopback_assigner[self.__loopback_pos]

                rnode.addStartCommand('ip li add dummy0 type dummy')
                rnode.addStartCommand('ip li set dummy0 up')
                rnode.addStartCommand('ip addr add {}/32 dev dummy0'.format(lbaddr))
                rnode.setLoopbackAddress(lbaddr)
                self.__loopback_pos += 1

                self._log("Bootstraping bird.conf for AS{} Router {}...".format(scope, name))

                self.__installBird(rnode)

                r_ifaces = rnode.getInterfaces()
                assert len(r_ifaces) > 0, "router node {}/{} has no interfaces".format(rs_node.getAsn(), rs_node.getName())

                ifaces = ''
                has_localnet = False

                for iface in r_ifaces:
                    net = iface.getNet()
                    if net in self.__direct_nets:
                        has_localnet = True
                        ifaces += RoutingFileTemplates["rnode_bird_direct_interface"].format(
                            interfaceName = net.getName()
                        )

                rnode.setFile("/etc/bird/bird.conf", RoutingFileTemplates["rnode_bird"].format(
                    routerId = rnode.getLoopbackAddress()
                ))

                rnode.addStartCommand('[ ! -d /run/bird ] && mkdir /run/bird')
                rnode.addStartCommand('bird -d', True)
                
                if has_localnet: rnode.addProtocol('direct', 'local_nets', RoutingFileTemplates['rnode_bird_direct'].format(interfaces = ifaces))

            if type == 'hnode':
                hnode: Node = obj
                hifaces: List[Interface] = hnode.getInterfaces()
                assert len(hifaces) == 1, 'Host {} in as{} has != 1 interfaces'.format(name, scope)
                hif = hifaces[0]
                hnet: Network = hif.getNet()
                rif: Interface = None

                cur_scope = ScopedRegistry(scope)
                for router in cur_scope.getByType('rnode'):
                    if rif != None: break
                    for riface in router.getInterfaces():
                        if riface.getNet() == hnet:
                            rif = riface
                            break
                
                assert rif != None, 'Host {} in as{} in network {}: no router'.format(name, scope, hnet.getName())
                self._log("Setting default route for host {} ({}) to router {}".format(name, hif.getAddress(), rif.getAddress()))
                hnode.addStartCommand('ip rou del default 2> /dev/null')
                hnode.addStartCommand('ip route add default via {} dev {}'.format(rif.getAddress(), rif.getNet().getName()))

    def addDirect(self, net: Network):
        """!
        @brief Add a network to "direct" protcol.

        Mark network as "direct" candidate. All router nodes connected to this
        network will add the interface attaches to this network to their
        "direct" protocol block. The "direct" protocol will be attached to the
        "t_direct" table.

        @param net network.
        @throws AssertionError if try to add non-AS network as direct.
        """
        assert net.getType() != NetworkType.InternetExchange, 'cannot set IX network {} as direct network.'.format(net.getName())
        self.__direct_nets.add(net)

    def addDirectByName(self, asn: int, netname: str):
        """!
        @brief Add a network to "direct" protcol by name.

        Mark network as "direct" candidate. All router nodes connected to this
        network will add the interface attaches to this network to their
        "direct" protocol block.

        @param asn ASN.
        @param netname network name.
        @throws AssertionError if net not exist.
        @throws AssertionError if try to add non-AS network as direct.
        """
        self.addDirect(self.__reg.get(str(asn), 'net', netname))

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'RoutingLayer:\n'

        indent += 4
        out += ' ' * indent

        out += 'Direct Networks:\n'

        indent += 4

        for net in self.__direct_nets:
            out += ' ' * indent
            (scope, _, netname) = net.getRegistryInfo()
            out += 'as{}/{}\n'.format(scope, netname)

        return out


from .Service import Service, Server
from .DomainNameService import DomainNameService, DomainNameServer, Zone
from .Reality import Reality
from seedsim.core import Node, ScopedRegistry, Registry, Network
from typing import List, Tuple
from ipaddress import IPv4Network

class CyrmuIpOriginServer(Server):
    """!
    @brief Cymru's IP info service server.
    """

    __node: Node

    def __init__(self, node: Node):
        """!
        @brief CyrmuIpOriginServer constructor.

        @param node node to install on.
        """
        self.__node = node

    def getNode(self) -> Node:
        """!
        @brief get node.

        @returns node.
        """
        return self.__node

class CyrmuIpOriginService(Service):
    """!
    @brief Cymru's IP info service.

    Cymru's IP info service is used by various traceroute utilities to map IP
    address to ASN (using DNS). This service loads the prefix list within the
    simulation and creates ASN mappings for them, so with proper local DNS
    configured, nodes can see the ASN when doing traceroute. 

    This layer hosts the domain cymru.com.
    """

    __servers: List[CyrmuIpOriginServer]
    __records: List[str]
    __reg = ScopedRegistry('seedsim')

    def __init__(self):
        """!
        @brief CyrmuIpOriginService constructor
        """
        self.__records = []
        self.__servers = []
        self.addDependency('DomainNameService', True, True)
        self.addDependency('Base', False, False)

    def getName(self) -> str:
        return 'CyrmuIpOriginService'

    def addMapping(self, prefix: str, asn: int):
        """!
        @brief Add new prefix -> asn mapping.

        @param prefix prefix.
        @param asn asn.

        @throws AssertionError if prefix invalid.
        """
        [pfx, cidr] = prefix.split('/')
        cidr = int(cidr)
        assert cidr <= 24, 'Invalid prefix.'
        prefix = IPv4Network(prefix)

        sub_cidr = 24
        num_8s = 3

        if cidr >= 0:
            sub_cidr = 8
            num_8s = 1

        if cidr >= 9:
            sub_cidr = 16
            num_8s = 2

        if cidr >= 17:
            sub_cidr = 24
            num_8s = 3

        for net in prefix.subnets(new_prefix = sub_cidr):
            record = '*.'
            record += '.'.join(reversed(str(net).split('.')[0:3]))
            record += '.origin.asn TXT "{} | {} | ZZ | SEED | 0000-00-00"'.format(asn, net)
            self.__records.append(record)

    def _doInstall(self, node: Node) -> CyrmuIpOriginServer: 
        server = CyrmuIpOriginServer(node)
        self.__servers.append(server)

    def onRender(self):
        mappings: List[Tuple[str, str]] = []

        if self.__reg.has('layer', 'Reality'):
            real: Reality = self.__reg.get('layer', 'Reality')
            for router in real.getRealWorldRouters():
                (asn, _, name) = router.getRegistryInfo()
                asn = int(asn)
                self._log('Collecting real-world route info on as{}/{}...'.format(asn, name))
                for prefix in router.getRealWorldRoutes():
                    mappings.append((prefix, asn))
        
        self._log('Collecting all networks in the simulation...')
        for regobj in (Registry()).getAll().items():
            [(asn, type, name), obj] = regobj
            if type != 'net': continue
            net: Network = obj
            if asn == 'ix': asn = name.replace('ix', '')
            mappings.append((net.getPrefix(), int(asn)))

        for mapping in mappings:
            (prefix, asn) = mapping
            self.addMapping(str(prefix), asn)

        self._log('Creating "cymru.com." zone...')
        dns: DomainNameService = self.__reg.get('layer', 'DomainNameService')
        zone = dns.getZone('cymru.com.')

        self._log('Setting up "cymru.com." server nodes...')
        for server in self.__servers:
            dns.hostZoneOn('cymru.com.', server.getNode())

        self._log('Adding mappings...')
        for record in self.__records:
            zone.addRecord(record)

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'CyrmuIpOriginService:\n'
        
        indent += 4
        out += ' ' * indent
        out += 'Installed on:\n'

        indent += 4
        for server in self.__servers:
            (asn, _, name) = server.getNode().getRegistryInfo()
            out += ' ' * indent
            out += 'as{}/{}\n'.format(asn, name)

        return out




from .Layer import Layer
from .Base import Base
from .Ospf import Ospf
from seedsim.core import Registry, ScopedRegistry, Node, Interface, Graphable
from seedsim.core.enums import NetworkType
from typing import List, Set, Dict

IbgpFileTemplates: Dict[str, str] = {}

IbgpFileTemplates['ibgp_peer'] = '''
    ipv4 {{
        table t_bgp;
        import all;
        export all;
        igp table t_ospf;
    }};
    local {localAddress} as {asn};
    neighbor {peerAddress} as {asn};
'''

class Ibgp(Layer, Graphable):
    """!
    @brief The Ibgp (iBGP) layer.

    @todo bird2

    This layer automatically setup full mesh peering between routers within AS.
    """
    __masked: Set[int] = set()
    __reg = Registry()

    def __init__(self):
        """!
        @brief Ibgp (iBGP) layer constructor.
        """
        Graphable.__init__(self)
        self.__masked = set()
        self.addDependency('Ospf', False, False)

    def getName(self) -> str:
        return 'Ibgp'

    def mask(self, asn: int):
        """!
        @brief Mask an AS.

        By default, Ibgp layer will add iBGP peering for all ASes. Use this
        method to mask an AS and disable iBGP.

        @param asn AS to mask.
        """
        self.__masked.add(asn)

    def __findFirstUnmasked(self, node: Node) -> Interface:
        """!
        @brief find first NIC on the node that is connected to a unmasked
        internal network.

        @param node target node.

        @returns interface if found, None if not found.
        """
        for iface in node.getInterfaces():
            ospf: Ospf = self.__reg.get('seedsim', 'layer', 'Ospf')
            net = iface.getNet()
            if net.getType() == NetworkType.InternetExchange: continue
            if ospf.isMasked(net): continue
            return iface

        return None

    def onRender(self):
        base: Base = self.__reg.get('seedsim', 'layer', 'Base')
        for asn in base.getAsns():
            if asn in self.__masked: continue

            self._log('setting up IBGP peering for as{}...'.format(asn))
            routers: List[Node] = ScopedRegistry(str(asn)).getByType('rnode')

            for local in routers:
                self._log('setting up IBGP peering on as{}/{}...'.format(asn, local.getName()))

                n = 1
                for remote in routers:
                    if local == remote: continue

                    laddr = local.getLoopbackAddress()
                    raddr = remote.getLoopbackAddress()
                    local.addTable('t_bgp')
                    local.addTablePipe('t_bgp')
                    local.addTablePipe('t_direct', 't_bgp')
                    local.addProtocol('bgp', 'ibgp{}'.format(n), IbgpFileTemplates['ibgp_peer'].format(
                        localAddress = laddr,
                        peerAddress = raddr,
                        asn = asn
                    ))

                    n += 1

                    self._log('adding peering: {} <-> {} (ibgp, as{})'.format(laddr, raddr, asn))

    def _doCreateGraphs(self):
        base: Base = self.__reg.get('seedsim', 'layer', 'Base')
        for asn in base.getAsns():
            if asn in self.__masked: continue
            asobj = base.getAutonomousSystem(asn)
            asobj.createGraphs()
            l2graph = asobj.getGraph('AS{}: Layer 2 Connections'.format(asn))
            ibgpgraph = self._addGraph('AS{}: iBGP sessions'.format(asn), False)
            ibgpgraph.copy(l2graph)
            for edge in ibgpgraph.edges:
                edge.style = 'dotted'

            rtrs = ScopedRegistry(str(asn)).getByType('rnode').copy()
            
            while len(rtrs) > 0:
                a = rtrs.pop()
                for b in rtrs:
                    ibgpgraph.addEdge('Router: {}'.format(a.getName()), 'Router: {}'.format(b.getName()), style = 'solid')
            

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'IbgpLayer:\n'

        indent += 4
        out += ' ' * indent
        out += 'Masked ASes:\n'

        indent += 4
        for asn in self.__masked:
            out += ' ' * indent
            out += '{}\n'.format(asn)

        return out




from .Layer import Layer
from seedsim.core import AutonomousSystem, InternetExchange, Printable, Registry, AddressAssignmentConstraint, Node, Graphable
from typing import Dict, List

BaseFileTemplates: Dict[str, str] = {}

BaseFileTemplates["interface_setup_script"] = """\
#!/bin/bash
cidr_to_net() {
    ipcalc -n "$1" | sed -E -n 's/^Network: +([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\/[0-9]{1,2}) +.*/\\1/p'
}

ip -j addr | jq -cr '.[]' | while read -r iface; do {
    ifname="`jq -cr '.ifname' <<< "$iface"`"
    jq -cr '.addr_info[]' <<< "$iface" | while read -r iaddr; do {
        addr="`jq -cr '"\(.local)/\(.prefixlen)"' <<< "$iaddr"`"
        net="`cidr_to_net "$addr"`"
        line="`grep "$net" < ifinfo.txt`"
        new_ifname="`cut -d: -f1 <<< "$line"`"
        latency="`cut -d: -f3 <<< "$line"`"
        bw="`cut -d: -f4 <<< "$line"`"
        [ "$bw" = 0 ] && bw=1000000000000
        loss="`cut -d: -f5 <<< "$line"`"
        [ ! -z "$new_ifname" ] && {
            ip li set "$ifname" down
            ip li set "$ifname" name "$new_ifname"
            ip li set "$new_ifname" up
            tc qdisc add dev "$new_ifname" root handle 1:0 tbf rate "${bw}bit" buffer 1000000 limit 1000
            tc qdisc add dev "$new_ifname" parent 1:0 handle 10: netem delay "${latency}ms" loss "${loss}%"
        }
    }; done
}; done
"""

class Base(Layer, Graphable):
    """!
    @brief The base layer.
    """

    __ases: Dict[int, AutonomousSystem]
    __ixes: Dict[int, InternetExchange]
    __reg = Registry()

    def __init__(self):
        """!
        @brief Base layer constructor.
        """
        Graphable.__init__(self)
        self.__ases = {}
        self.__ixes = {}

    def getName(self) -> str:
        return "Base"

    def onRender(self) -> None:
        for ((scope, type, name), obj) in self.__reg.getAll().items():

            if not (type == 'rs' or type == 'rnode' or type == 'hnode'):
                continue
            
            node: Node = obj

            ifinfo = ''
            for iface in node.getInterfaces():
                net = iface.getNet()
                [l, b, d] = iface.getLinkProperties()
                ifinfo += '{}:{}:{}:{}:{}\n'.format(net.getName(), net.getPrefix(), l, b, d)

            node.setFile('/ifinfo.txt', ifinfo)
            node.setFile('/interface_setup', BaseFileTemplates['interface_setup_script'])
            node.addStartCommand('chmod +x /interface_setup')
            node.addStartCommand('/interface_setup')

    def createAutonomousSystem(self, asn: int) -> AutonomousSystem:
        """!
        @brief Create a new AutonomousSystem.

        @param asn ASN of the new AS.
        @returns created AS.
        @throws AssertionError if asn exists.
        """
        assert asn not in self.__ases, "as{} already exist.".format(asn)
        self.__ases[asn] = AutonomousSystem(asn)
        return self.__ases[asn]

    def getAutonomousSystem(self, asn: int) -> AutonomousSystem:
        """!
        @brief Create an existing AutonomousSystem.

        @param asn ASN of the AS.
        @returns AS.
        @throws AssertionError if asn does not exist.
        """
        assert asn in self.__ases, "as{} does not exist.".format(asn)
        return self.__ases[asn]

    def createInternetExchange(self, asn: int, prefix: str = "auto", aac: AddressAssignmentConstraint = None) -> InternetExchange:
        """!
        @brief Create a new InternetExchange.

        @param asn ASN of the new IX.
        @param prefix (optional) prefix of the IX peering LAN.
        @param aac (optional) Address assigment constraint.
        @returns created IX.
        @throws AssertionError if IX exists.
        """
        assert asn not in self.__ixes, "ix{} already exist.".format(asn)
        self.__ixes[asn] = InternetExchange(asn, prefix, aac)
        return self.__ixes[asn]

    def getInternetExchange(self, asn: int) -> InternetExchange:
        """!
        @brief Get an existing InternetExchange.

        @param asn ASN of the IX.
        @returns InternetExchange.
        @throws AssertionError if ix does not exist.
        """
        assert asn in self.__ixes, "ix{} does not exist.".format(asn)
        return self.__ixes[asn]

    def getAsns(self) -> List[int]:
        """!
        @brief Get list of ASNs.

        @returns List of ASNs.
        """
        return self.__ases.keys()

    def getIxIds(self) -> List[int]:
        """!
        @brief Get list of IX IDs.

        @returns List of IX IDs.
        """
        return self.__ixes.keys()

    def _doCreateGraphs(self):
        graph = self._addGraph('Layer 2 Connections', False)
        for asobj in self.__ases.values():
            asobj.createGraphs()
            asgraph = asobj.getGraph('AS{}: Layer 2 Connections'.format(asobj.getAsn()))
            graph.copy(asgraph)

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'BaseLayer:\n'

        indent += 4
        out += ' ' * indent
        out += 'AutonomousSystems:\n'
        for _as in self.__ases.values():
            out += _as.print(indent + 4)

        out += ' ' * indent
        out += 'InternetExchanges:\n'
        for _as in self.__ixes.values():
            out += _as.print(indent + 4)

        return out


from typing import Dict, Tuple, List
from .Printable import Printable

class Registrable(object):
    """!
    @brief The Registerable base class.

    Base class for all Registrable objects.
    """
    _rscope: str
    _rtype: str
    _rname: str
    _attrs: Dict[str, object]

    def __init__(self):
        """!
        @brief Registerable class constructor.
        """
        self.__scope = 'undefined'
        self._rtype = 'undefined'
        self._rname = 'undefined'

    def doRegister(self, scope: str, type: str, name: str):
        """!
        @brief Handle registration.

        @param scope scope.
        @param type type.
        @param name name.
        """
        self._rscope = scope
        self._rtype = type
        self._rname = name
        self._attrs = {}
    
    def getRegistryInfo(self) -> Tuple[str, str, str]:
        """!
        @brief Get registry info

        @returns Tuple of scope, type and name
        """
        return (self._rscope, self._rtype, self._rname)

    def getAttribute(self, name: str, default: object = None) -> object:
        """!
        @brief Get an attribute.

        @param name name of attribute.
        @param default value to set and return if name not exist.

        @returns value, or None if not exist.
        """
        if name not in self._attrs:
            if default != None:
                self.setAttribute(name, default)
                return self._attrs[name]
            return None
        return self._attrs[name]

    def setAttribute(self, name: str, value: object):
        """!
        @brief Set an attribute.

        @param name name of attribute.
        @param value value of attribute.
        """
        self._attrs[name] = value

    def hasAttribute(self, name: str) -> bool:
        """!
        @brief Check if an attribute exists.

        @param name name of attribute.
        
        @returns True if exist, False otherwise.
        """
        return name in self._attrs

class Registry(Printable):
    """!
    @brief The Registry class.

    Registry is the global container for all objects in the simulator.
    """

    __objects: Dict[Tuple[str, str, str], Registrable] = {}

    def register(self, scope: str, type: str, name: str, obj: Registrable) -> Registrable:
        """!
        @brief Register an object.

        @param scope scope of the object (e.g., asn).
        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @param obj target object.
        @returns registered object
        @throws AssertionError if name exists.
        """
        assert (scope, type, name) not in self.__objects, 'object with name {} already exist.'.format(name)
        obj.doRegister(scope, type, name)
        self.__objects[(scope, type, name)] = obj
        return self.__objects[(scope, type, name)]

    def get(self, scope: str, type: str, name: str) -> Registrable:
        """!
        @brief Retrive an object with name.

        @param scope scope of the object (e.g., asn).
        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @throws AssertionError if name does not exist.
        @returns object.
        """
        assert (scope, type, name) in self.__objects, 'object with name {} does not exist.'.format(name)
        return self.__objects[(scope, type, name)]

    def has(self, scope: str, type: str, name: str) -> bool:
        """!
        @brief Test if an object exist.

        @param scope scope of the object (e.g., asn).
        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @returns True if exist, False otherwise.
        """
        return (scope, type, name) in self.__objects

    def getByType(self, scope: str, type: str) -> List[Registrable]:
        """!
        @brief Retrive objects with type.

        @param scope scope of the object (e.g., asn).
        @param type type of the object (e.g., net/node)
        @returns objects.
        """
        rslt: List[Registrable] = []

        for key, obj in self.__objects.items():
            (s, t, _) = key
            if s == scope and t == type: rslt.append(obj)

        return rslt

    def getAll(self) -> Dict[Tuple[str, str, str], Registrable]:
        """!
        @brief Get all objects in the Global Registry.

        @returns dictionary, where keys in tuple (scope, type, name) and value
        is object
        """
        return self.__objects

    def getByScope(self, scope: str) -> List[Registrable]:
        """!
        @brief Retrive objects with scope.

        @param scope scope of the object (e.g., asn).
        @returns objects.
        """
        rslt: List[Registrable] = []

        for key, obj in self.__objects.items():
            (s, _, _) = key
            if s == scope: rslt.append(obj)

        return rslt
    
    def print(self, indent: int):
        out = (' ' * indent) + 'Registry:\n'
        indent += 4
        for keys, val in self.__objects.items():
            [scope, type, name] = keys
            out += (' ' * indent) + 'Object {}/{}/{}:\n'.format(scope, type, name)
            out += val.print(indent + 4)

        return out

class ScopedRegistry(Registry):
    """!
    @brief Scoped Registry class.

    Scoped wrapper for Registry class.
    """

    __reg = Registry()
    __scope: str

    def __init__(self, scope: str):
        """!
        @brief Scoped Registry ctor.

        @param scope scope to bind to.
        """
        self.__scope = scope

    def register(self, type: str, name: str, obj: Registrable) -> Registrable:
        """!
        @brief Register an object.

        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @param obj target object.
        @returns registered object
        @throws AssertionError if name exists.
        """
        return self.__reg.register(self.__scope, type, name, obj)

    def get(self, type: str, name: str) -> object:
        """!
        @brief Retrive an object with name.

        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @throws AssertionError if name does not exist.
        @returns object.
        """
        return self.__reg.get(self.__scope, type, name)

    def has(self, type: str, name: str) -> bool:
        """!
        @brief Test if an object exist.

        @param type type of the object (e.g., net/node)
        @param name name of the object.
        @returns True if exist, False otherwise.
        """
        return self.__reg.has(self.__scope, type, name)

    def getByType(self, type: str) -> List[Registrable]:
        """!
        @brief Retrive objects with type.

        @param type type of the object (e.g., net/node)
        @returns objects.
        """
        return self.__reg.getByType(self.__scope, type)


from ipaddress import IPv4Network, IPv4Address
from .Printable import Printable
from .enums import NetworkType, NodeRole
from .Registry import Registrable
from .AddressAssignmentConstraint import AddressAssignmentConstraint
from typing import Generator, Dict, Tuple, List

class Network(Printable, Registrable):
    """!
    @brief The network class.

    This class represents a network.
    """
    __type: NetworkType
    __prefix: IPv4Network
    __name: str
    __scope: str
    __aac: AddressAssignmentConstraint
    __assigners: Dict[NodeRole, Generator[int, None, None]]

    __connected_nodes: List['Node']

    __d_latency: int       # in ms
    __d_bandwidth: int     # in bps
    __d_drop: float        # percentage

    __mtu: int

    def __init__(self, name: str, type: NetworkType, prefix: IPv4Network, aac: AddressAssignmentConstraint = None):
        """!
        @brief Network constructor.

        @param name name of the network. Note that this is considered a "local"
        name. Networks can have the same name, as long as they are in different
        contexts (i.e., different AS).
        @param type type of the network.
        @param prefix prefix of the network.
        @param aac (optional) AddressAssignmentConstraint to use.
        """
        self.__name = name
        self.__type = type
        self.__prefix = prefix
        self.__aac = aac if aac != None else AddressAssignmentConstraint()
        self.__assigners = {}

        self.__connected_nodes = []

        self.__assigners[NodeRole.Router] = self.__aac.getOffsetGenerator(NodeRole.Router)
        self.__assigners[NodeRole.Host] = self.__aac.getOffsetGenerator(NodeRole.Host)

        self.__d_latency = 0
        self.__d_bandwidth = 0
        self.__d_drop = 0

        self.__mtu = 1500

    def setMtu(self, mtu: int):
        """!
        @brief Set MTU of this network.

        @param mtu MTU.
        """
        self.__mtu = mtu

    def getMtu(self) -> int:
        """!
        @brief Get MTU of this network.

        @returns mtu.
        """
        return self.__mtu

    def setDefaultLinkProperties(self, latency: int = 0, bandwidth: int = 0, packetDrop: float = 0):
        """!
        @brief Set default link properties of interfaces attached to the network.

        @param latency (optional) latency to add to the link in ms, default 0. Note that this will be
        apply on all interfaces, meaning the rtt between two hosts will be 2 * latency.
        @param bandwidth (optional) egress bandwidth of the link in bps, 0 for unlimited, default 0.
        @param packetDrop (optional) link packet drop as percentage, 0 for unlimited, default 0.
        """
        assert latency >= 0, 'invalid latency'
        assert bandwidth >= 0, 'invalid bandwidth'
        assert packetDrop >= 0 and packetDrop <= 100, 'invalid packet drop'

        self.__d_latency = latency
        self.__d_bandwidth = bandwidth
        self.__d_drop = packetDrop

    def getDefaultLinkProperties(self) -> Tuple[int, int, int]:
        """!
        @brief Get default link properties.

        @returns tuple (latency, bandwidth, packet drop)
        """
        return (self.__d_latency, self.__d_bandwidth, self.__d_drop)

    def getName(self) -> str:
        """!
        @brief Get name of this network.

        @returns name.
        """
        return self.__name

    def getType(self) -> NetworkType:
        """!
        @brief Get type of this network.

        @returns type.
        """
        return self.__type

    def getPrefix(self) -> IPv4Network:
        """!
        @brief Get prefix of this network.

        @returns prefix.
        """
        return self.__prefix

    def assign(self, nodeRole: NodeRole, asn: int = -1) -> IPv4Address:
        """!
        @brief Assign IP for interface.

        @param nodeRole role of the node getting this assigment.
        @param asn optional. If interface type is InternetExchange, the asn for
        IP address mapping.
        """
        assert not (nodeRole == nodeRole.Host and self.__type == NetworkType.InternetExchange), 'trying to assign IX netwotk to non-router node'

        if self.__type == NetworkType.InternetExchange: return self.__prefix[self.__aac.mapIxAddress(asn)]
        return self.__prefix[next(self.__assigners[nodeRole])]

    def associate(self, node: 'Node'):
        """!
        @brief Associate the node with network.

        @param node node.
        """
        self.__connected_nodes.append(node)

    def getAssociations(self) -> List['Node']:
        """!
        @brief Get list of assoicated nodes.

        @returns list of nodes.
        """
        return self.__connected_nodes

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'Network {} ({}):\n'.format(self.__name, self.__type)

        indent += 4
        out += ' ' * indent
        out += 'Prefix: {}\n'.format(self.__prefix)
        out += self.__aac.print(indent)

        return out


from .Layer import Layer
from seedsim.core import Node, Printable, ScopedRegistry
from seedsim.core.enums import NodeRole
from typing import Dict, List

class Server(Printable):
    """!
    @brief Server class.

    The Server class is the handler for installed services.
    """

class Service(Layer):
    """!
    @brief Service base class.

    The base class for all Services.
    """

    def _doInstall(self, node: Node) -> Server:
        """!
        @brief Install the service on node.

        @param node node to install the service on.
        """
        raise NotImplementedError('_doInstall not implemented')

    def _getStorage(self) -> Dict[str, object]:
        """!
        @brief Get a node-specific service storage.

        @returns dict for storage.
        """

    def getConflicts(self) -> List[str]:
        """!
        @brief Get a list of conflicting services.

        Override to change.

        @return list of service names.
        """
        return []

    def installOn(self, node: Node) -> Server:
        """!
        @brief Install the service on given node.

        @param node node to install the service on.

        @returns Handler of the installed service.
        @throws AssertionError if node is not host node.
        """
        assert node.getRole() == NodeRole.Host, 'node as{}/{} is not a host node'.format(node.getAsn(), node.getName())
        servicesdb: Dict = node.getAttribute('services', {})

        for (name, service_info) in servicesdb.items():
            service: Service = service_info['__self']
            assert name not in self.getConflicts(), '{} conflict with {} on as{}/{}.'.format(self.getName(), service.getName(), node.getAsn(), node.getName())
            assert self.getName() not in service.getConflicts(), '{} conflict with {} on as{}/{}.'.format(self.getName(), service.getName(), node.getAsn(), node.getName())

        m_name = self.getName()
        if m_name not in servicesdb:
            servicesdb[m_name] = {
                '__self': self
            }

        return self._doInstall(node)

    def installOnAll(self, asn: int) -> Dict[int, Dict[str, Server]]:
        """!
        @brief install the service on all host nodes in the given AS.

        @param asn ASN.

        @returns Dict of Dict, the inner dict is Dict[Node name, Server] and the
        outer dict is Dict[Asn, inner Dict].
        """
        scope = ScopedRegistry(str(asn))
        for host in scope.getByType('hnode'):
            self.installOn(host)




from .Service import Service, Server
from seedsim.core import Node
from seedsim.core.enums import NodeRole
from typing import List, Dict

WebServerFileTemplates: Dict[str, str] = {}

WebServerFileTemplates['nginx_site'] = '''\
server {{
    listen {port};
    root /var/www/html;
    index index.html;
    server_name _;
    location / {{
        try_files $uri $uri/ =404;
    }}
}}
'''

class WebServer(Server):
    """!
    @brief The WebServer class.
    """

    __node: Node
    __port: int
    __index: str

    def __init__(self, node: Node):
        """!
        @brief WebServer constructor.

        @param node node.
        """
        asn = node.getAsn()
        self.__node = node
        self.__port = 80
        self.__index = '<h1>Web Server node {} at AS{}</h1>'.format(node.getName(), node.getAsn())
        

    def setPort(self, port: int):
        """!
        @brief Set HTTP port.

        @param port port.
        """
        ## ! todo
        self.__port = port

    def setIndexContent(self, content: str):
        """!
        @brief Set content of index.html.

        @param content content.
        """
        self.__content = content
    
    def install(self):
        """!
        @brief Install the service.
        """
        self.__node.addSoftware('nginx-light')
        self.__node.setFile('/var/www/html/index.html', self.__index)
        self.__node.setFile('/etc/nginx/sites-available/default', WebServerFileTemplates['nginx_site'].format(port = self.__port))
        self.__node.addStartCommand('service nginx start')
        
    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'Server: as{}/{}\n'.format(self.__node.getAsn(), self.__node.getName())

        return out

class WebService(Service):
    """!
    @brief The WebService class.
    """

    __servers: List[WebServer]

    def __init__(self):
        """!
        @brief WebService constructor.
        """
        self.__servers = []
        self.addDependency('Base', False, False)

    def getName(self) -> str:
        return 'WebService'

    def _doInstall(self, node: Node) -> WebServer:
        """!
        @brief Install the web service on given node.

        @param node node to install the web service on.

        @returns Handler of the installed web service.
        @throws AssertionError if node is not host node.
        """
        server: WebServer = node.getAttribute('__web_service_server')
        if server != None: return server
        server = WebServer(node)
        self.__servers.append(server)
        node.setAttribute('__web_service_server', server)
        return server

    def onRender(self):
        for server in self.__servers:
            server.install()

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'WebServiceLayer:\n'

        indent += 4
        out += ' ' * indent

        out += 'Installed Nodes:\n'
 
        for server in self.__servers:
            out += server.print(indent + 4) 

        return out       


from enum import Enum

class NetworkType(Enum):
    """!
    @brief Network types enum.
    """

    ## Public internet exchange network
    InternetExchange = "Internet Exchange Network"

    ## Private links netwroks. OSPF routers and IBGP runs on this type of
    #  networks.
    Local = "Local Network"

    ## Host networks. "Content" (i.e., web server, dns server) runs on this type
    #  of networks.
    Host = "Content Network"

class NodeRole(Enum):
    """!
    @brief Node roles enum.
    """

    ## Host node.
    Host = "Host"

    ## Router node.
    Router = "Router"

    ## Route served node.
    RouteServer = "Route Server"


class Printable:
    """!
    @brief Printable class.

    Implement this class for indentable print.
    """

    def print(self, indentation: int = 0) -> str:
        """!
        @brief get printable string.

        @param indentation indentation.

        @returns printable string.
        """

        raise NotImplementedError("print not implemented.")

    def __str__(self) -> str:
        """!
        @brief convert to string.

        alias to print(0).
        """
        return self.print(0)


from .Layer import Layer
from .Routing import Router
from seedsim.core import Registry, ScopedRegistry, Network, Node, Interface, Graphable
from typing import Tuple, List, Dict

EbgpFileTemplates: Dict[str, str] = {}

EbgpFileTemplates["rs_bird_peer"] =  """
    ipv4 {{
        import all;
        export all;
    }};
    rs client;
    local {localAddress} as {localAsn};
    neighbor {peerAddress} as {peerAsn};
"""

EbgpFileTemplates["rnode_bird_peer"] = """
    ipv4 {{
        table t_bgp;
        import {importFilter};
        export {exportFilter};
        next hop self;
    }};
    local {localAddress} as {localAsn};
    neighbor {peerAddress} as {peerAsn};
"""

class Ebgp(Layer, Graphable):
    """!
    @brief The Ebgp (eBGP) layer.

    This layer enable eBGP peering in InternetExchange.
    """

    __peerings: List[Tuple[int, int, int]]
    __rs_peers: List[Tuple[int, int]]

    def __init__(self):
        """!
        @brief Ebgp layer constructor.
        """
        Graphable.__init__(self)
        self.__peerings = []
        self.__rs_peers = []
        self.addDependency('Routing', False, False)
    
    def getName(self) -> str:
        return "Ebgp"

    def addPrivatePeering(self, ix: int, a: int, b: int):
        """!
        @brief Setup private peering between two ASes in IX.

        @param ix IXP id.
        @param a First ASN.
        @param b Second ASN.

        @throws AssertionError if peering already exist.
        """
        assert (ix, a, b) not in self.__peerings, '{} <-> {} already peered at IX{}'.format(a, b, ix)
        assert (ix, b, a) not in self.__peerings, '{} <-> {} already peered at IX{}'.format(b, a, ix)

        self.__peerings.append((ix, a, b))

    def addRsPeer(self, ix: int, peer: int):
        """!
        @brief Setup RS peering for an AS.

        @param ix IXP id.
        @param peer Participant ASN.

        @throws AssertionError if peering already exist.
        """
        assert (ix, peer) not in self.__rs_peers, '{} already peered with RS at IX{}'.format(peer, ix)

        self.__rs_peers.append((ix, peer))

    def onRender(self) -> None:
        for (ix, peer) in self.__rs_peers:
            ix_reg = ScopedRegistry('ix')
            p_reg = ScopedRegistry(str(peer))

            ix_net: Network = ix_reg.get('net', 'ix{}'.format(ix))
            ix_rs: Router = ix_reg.get('rs', 'ix{}'.format(ix))
            rs_ifs = ix_rs.getInterfaces()
            assert len(rs_ifs) == 1, '??? ix{} rs has {} interfaces.'.format(ix, len(rs_ifs))
            rs_if = rs_ifs[0]

            p_rnodes: List[Router] = p_reg.getByType('rnode')
            p_ixnode: Router = None
            p_ixif: Interface = None
            for node in p_rnodes:
                if p_ixnode != None: break
                for iface in node.getInterfaces():
                    if iface.getNet() == ix_net:
                        p_ixnode = node
                        p_ixif = iface
                        break

            assert p_ixnode != None, 'cannot resolve peering: as{} not in ix{}'.format(a, ix)
            self._log("adding peering: {} as {} (RS) <-> {} as {}".format(rs_if.getAddress(), ix, p_ixif.getAddress(), peer))

            ix_rs.addProtocol('bgp', 'as{}'.format(peer), EbgpFileTemplates["rs_bird_peer"].format(
                localAddress = rs_if.getAddress(),
                localAsn = ix,
                peerAddress = p_ixif.getAddress(),
                peerAsn = peer
            )) 

            p_ixnode.addTable('t_bgp')
            p_ixnode.addTablePipe('t_bgp')
            p_ixnode.addTablePipe('t_direct', 't_bgp')
            p_ixnode.addProtocol('bgp', 'rs{}'.format(ix), EbgpFileTemplates["rnode_bird_peer"].format(
                localAddress = p_ixif.getAddress(),
                localAsn = peer,
                peerAddress = rs_if.getAddress(),
                peerAsn = ix,
                exportFilter = "all", # !! todo
                importFilter = "all"
            )) 

        for (ix, a, b) in self.__peerings:
            ix_reg = ScopedRegistry('ix')
            a_reg = ScopedRegistry(str(a))
            b_reg = ScopedRegistry(str(b))

            ix_net: Network = ix_reg.get('net', 'ix{}'.format(ix))
            a_rnodes: List[Router] = a_reg.getByType('rnode')
            b_rnodes: List[Router] = b_reg.getByType('rnode')

            a_ixnode: Router = None
            a_ixif: Interface = None
            for node in a_rnodes:
                if a_ixnode != None: break
                for iface in node.getInterfaces():
                    if iface.getNet() == ix_net:
                        a_ixnode = node
                        a_ixif = iface
                        break
            
            assert a_ixnode != None, 'cannot resolve peering: as{} not in ix{}'.format(a, ix)

            b_ixnode: Router = None
            b_ixif: Interface = None
            for node in b_rnodes:
                if b_ixnode != None: break
                for iface in node.getInterfaces():
                    if iface.getNet() == ix_net:
                        b_ixnode = node
                        b_ixif = iface
                        break
            
            assert b_ixnode != None, 'cannot resolve peering: as{} not in ix{}'.format(b, ix)

            self._log("adding peering: {} as {} <-> {} as {}".format(a_ixif.getAddress(), a, b_ixif.getAddress(), b))

            # @todo import/export filter?
            a_ixnode.addTable('t_bgp')
            a_ixnode.addTablePipe('t_bgp')
            a_ixnode.addTablePipe('t_direct', 't_bgp')
            a_ixnode.addProtocol('bgp', 'as{}'.format(b), EbgpFileTemplates["rnode_bird_peer"].format(
                localAddress = a_ixif.getAddress(),
                localAsn = a,
                peerAddress = b_ixif.getAddress(),
                peerAsn = b,
                exportFilter = "all",
                importFilter = "all"
            ))

            # @todo import/export filter?
            b_ixnode.addTable('t_bgp')
            b_ixnode.addTablePipe('t_bgp')
            b_ixnode.addTablePipe('t_direct', 't_bgp')
            b_ixnode.addProtocol('bgp', 'as{}'.format(a), EbgpFileTemplates["rnode_bird_peer"].format(
                localAddress = b_ixif.getAddress(),
                localAsn = b,
                peerAddress = a_ixif.getAddress(),
                peerAsn = a,
                exportFilter = "all",
                importFilter = "all"
            ))

    def _doCreateGraphs(self):
        # creates the following:
        # - ebgp peering, all ASes in one graph
        # - ebgp peering, one for each ix
        # mlpa peer (i.e., via rs): dashed line
        # private peer: solid line

        full_graph = self._addGraph('All Peering Sessions', False)

        ix_list = set()
        for (i, _) in self.__rs_peers: ix_list.add(i)
        for (i, _, _) in self.__peerings: ix_list.add(i)
        for ix in ix_list:
            self._log('Creating RS peering sessions graph for IX{}...'.format(ix))
            ix_graph = self._addGraph('IX{} Peering Sessions'.format(ix), False)

            mesh_ases = set()
            
            for (i, a) in self.__rs_peers:
                if i == ix: mesh_ases.add(a)

            while len(mesh_ases) > 0:
                a = mesh_ases.pop()
                if not full_graph.hasVertex('AS{}'.format(a), 'IX{}'.format(i)):
                    full_graph.addVertex('AS{}'.format(a), 'IX{}'.format(i))
                if not ix_graph.hasVertex('AS{}'.format(a), 'IX{}'.format(i)):
                    ix_graph.addVertex('AS{}'.format(a), 'IX{}'.format(i))
                for b in mesh_ases:
                    if not full_graph.hasVertex('AS{}'.format(b), 'IX{}'.format(i)):
                        full_graph.addVertex('AS{}'.format(b), 'IX{}'.format(i))
                    if not ix_graph.hasVertex('AS{}'.format(b), 'IX{}'.format(i)):
                        ix_graph.addVertex('AS{}'.format(b), 'IX{}'.format(i))

                    full_graph.addEdge('AS{}'.format(a), 'AS{}'.format(b), 'IX{}'.format(i), 'IX{}'.format(i), style = 'dashed', )
                    ix_graph.addEdge('AS{}'.format(a), 'AS{}'.format(b), 'IX{}'.format(i), 'IX{}'.format(i), style = 'dashed')
                    
        for (i, a, b) in self.__peerings:
            self._log('Creating private peering sessions graph for IX{} AS{} <-> AS{}...'.format(i, a, b))

            ix_graph = self._addGraph('IX{} Peering Sessions'.format(i), False)

            if not full_graph.hasVertex('AS{}'.format(a), 'IX{}'.format(i)):
                full_graph.addVertex('AS{}'.format(a), 'IX{}'.format(i))
            if not ix_graph.hasVertex('AS{}'.format(a), 'IX{}'.format(i)):
                ix_graph.addVertex('AS{}'.format(a), 'IX{}'.format(i))

            if not full_graph.hasVertex('AS{}'.format(b), 'IX{}'.format(i)):
                full_graph.addVertex('AS{}'.format(b), 'IX{}'.format(i))
            if not ix_graph.hasVertex('AS{}'.format(b), 'IX{}'.format(i)):
                ix_graph.addVertex('AS{}'.format(b), 'IX{}'.format(i))

            full_graph.addEdge('AS{}'.format(a), 'AS{}'.format(b), 'IX{}'.format(i), 'IX{}'.format(i))
            ix_graph.addEdge('AS{}'.format(a), 'AS{}'.format(b), 'IX{}'.format(i), 'IX{}'.format(i))

        es = list(full_graph.vertices.values())
        while len(es) > 0:
            a = es.pop()
            for b in es:
                if a.name == b.name:
                    full_graph.addEdge(a.name, b.name, a.group, b.group, style = 'dotted')


    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'EbgpLayer:\n'

        indent += 4
        for (i, a) in self.__rs_peers:
            out += ' ' * indent
            out += 'IX{}: RS <-> AS{}\n'.format(i, a)

        for (i, a, b) in self.__peerings:
            out += ' ' * indent
            out += 'IX{}: AS{} <-> AS{}\n'.format(i, a, b)


        return out




from .Compiler import Compiler
from seedsim.core import ScopedRegistry, Node, Graphable
from typing import Dict
from unicodedata import normalize
import re


class Graphviz(Compiler):
    """!
    @brief Get all graphable object and graph them.

    """

    def __slugify(self, filename):
        return ''.join([c for c in filename if c.isalpha() or c.isdigit() or c == ' ']).rstrip()

    def getName(self) -> str:
        return 'Graphviz'

    def _doCompile(self):
        reg = ScopedRegistry('seedsim')
        for obj in reg.getByType('graph'):
            graphs: Graphable = obj
            graphs.createGraphs()
            for graph in graphs.getGraphs().values():
                print(graph.toGraphviz(), file=open('{}.dot'.format(self.__slugify(graph.name)), 'w'))


from seedsim.core import Printable, ScopedRegistry
from seedsim.layers import Layer
from typing import List, Tuple, Dict
from sys import stderr

class Renderer(Printable):
    """!
    @brief The Renderer.
    """

    __layers: Dict[str, Tuple[Layer, bool]]
    __reg = ScopedRegistry('seedsim')

    def __init__(self):
        """!
        @brief Renderer constructor.
        """
        self.__layers = {}

    def addLayer(self, layer: Layer):
        """!
        @brief Add a layer.

        @param layer layer to add.
        @throws AssertionError if layer already exist.
        """

        lname = layer.getName()
        assert lname not in self.__layers, 'layer {} already added.'.format(lname)
        self.__layers[lname] = (layer, False)
        self.__reg.register('layer', lname, layer)

    def __render(self, layerName: str, optional: bool):
        """!
        @brief Render a layer.
        
        @param layerName name of layer.
        @throws AssertionError if dependencies unmet 
        """
        self.__log('requesting render: {}'.format(layerName))

        if optional and layerName not in self.__layers:
            self.__log('{}: not found but is optional, skipping'.format(layerName))
            return

        assert layerName in self.__layers, 'Layer {} requried but missing'.format(layerName)

        (layer, done) = self.__layers[layerName]
        if done:
            self.__log('{}: already rendered, skipping'.format(layerName))
            return

        if layerName in Layer.dependencies:
            for (dep, opt) in Layer.dependencies[layerName]:
                self.__log('{}: requesting dependency render: {}'.format(layerName, dep))
                self.__render(dep, opt)

        self.__log('rendering {}...'.format(layerName))
        layer.onRender()
        self.__log('done: {}'.format(layerName))
        self.__layers[layerName] = (layer, True)

    def render(self):
        """!
        @brief Render.

        @throws AssertionError if dependencies unmet 
        """
        for layerName in self.__layers.keys():
            self.__render(layerName, False)

    def print(self, indent: int) -> str:
        out = ''
        for (k, (v, _)) in self.__layers.items():
            out += v.print(indent)

        return out

    def __log(self, message: str):
        print('== Renderer: {}'.format(message), file=stderr)


        



from .Layer import Layer
from .Ospf import Ospf
from .Ibgp import Ibgp
from .Routing import Router
from seedsim.core import Node, Registry, ScopedRegistry, Graphable
from seedsim.core.enums import NetworkType, NodeRole
from typing import List, Tuple, Dict

MplsFileTemplates: Dict[str, str] = {}

MplsFileTemplates['frr_start_script'] = """\
#!/bin/bash
mount -o remount rw /proc/sys 2> /dev/null
echo '1048575' > /proc/sys/net/mpls/platform_labels
while read -r iface; do echo '1' > "/proc/sys/net/mpls/conf/$iface/input"; done < mpls_ifaces.txt
sed -i 's/ldpd=no/ldpd=yes/' /etc/frr/daemons
sed -i 's/ospfd=no/ospfd=yes/' /etc/frr/daemons
service frr start
"""

MplsFileTemplates['frr_config'] = """\
router-id {loopbackAddress}
{ospfInterfaces}
mpls ldp
 address-family ipv4
  discovery transport-address {loopbackAddress}
{ldpInterfaces}
 exit-address-family
router ospf
 redistribute connected
"""

MplsFileTemplates['frr_config_ldp_iface'] = """\
  interface {interface}
"""

# todo: make configurable hello/dead
MplsFileTemplates['frr_config_ospf_iface'] = """\
interface {interface}
 ip ospf area 0
 ip ospf dead-interval minimal hello-multiplier 2
"""

MplsFileTemplates['bird_ibgp_peer'] = '''
    ipv4 {{
        table t_bgp;
        import all;
        export all;
        igp table master4;
    }};
    local {localAddress} as {asn};
    neighbor {peerAddress} as {asn};
'''


class Mpls(Layer, Graphable):
    """!
    @brief The Mpls (MPLS) layer.

    This layer is a replacement for the iBGP full mesh setup for the transit
    provider's internal network. Instead of the traditional IP network, which
    requires every hop to have a copy of the full table, MPLS allows non-edge
    hops to hold only the MPLS forwarding table, which negated the need for the
    full table.

    MPLS layer will setup iBGP, LDP, and OSPF. FRRouting will do LDP and OSPF,
    and BIRD will still do BGP. When installed, the MPLS layer will treat all
    nodes with (1) no connection to IX and (2) no connection to a network with
    at least one host node as non-edge nodes and will not put it as part of the
    iBGP mesh network.
    
    The MPLS layer requires kernel modules support. Make sure you load the
    following modules:

    - mpls_router
    - mpls_iptunnel
    - mpls_gso

    Node with MPLS enabled will be privileged. This means the container
    potentially have full control over the docker host. Be careful when exposing
    the node to the public.
    """

    __reg = Registry()
    __additional_edges: List[Node]
    __enabled: List[int]

    def __init__(self):
        """!
        @brief Mpls layer constructor.
        """
        Graphable.__init__(self)
        self.__additional_edges = []
        self.__enabled = []

        # they are not really "dependency," we just need them to render after
        # us, in case we need to setup masks.
        self.addDependency('Ospf', True, True)
        self.addDependency('Ibgp', True, True)

        self.addDependency('Routing', False, False)

    def getName(self) -> str:
        return 'Mpls'

    def markAsEdge(self, node: Node):
        """!
        @brief Mark a node as edge node.

        By default, only nodes with connection to IX, or connection to a network
        with at least one host node, will be considered an edge router and be
        included in the iBGP mesh. Use this method to mark a node as edge
        manually.

        @param node node
        """
        self.__additional_edges.append(node)

    def enableOn(self, asn: int):
        """!
        @brief Use MPLS in an AS.

        MPLS is not enabled by default. Use this method to enable MPLS for an
        AS. This also automatically setup masks for OSPF and IBGP layer if they
        exist.

        @param asn ASN.
        """
        self.__enabled.append(asn)

    def __getEdgeNodes(self, scope: ScopedRegistry) -> Tuple[List[Node], List[Node]]:
        """!
        @brief Helper tool - get list of routers (edge, non-edge) of an AS.

        @param scope scope.
        """
        enodes: List[Node] = []
        nodes: List[Node] = []

        for obj in scope.getByType('rnode'):
            node: Node = obj

            if node in self.__additional_edges:
                enodes.append(node)
                continue

            is_edge = False
            for iface in node.getInterfaces():
                net = iface.getNet()
                if net.getType() == NetworkType.InternetExchange:
                    is_edge = True
                    break
                if True in (node.getRole() == NodeRole.Host for node in net.getAssociations()):
                    is_edge = True
                    break

            if is_edge: enodes.append(node)
            else: nodes.append(node)

        return (enodes, nodes)

    def __setUpLdpOspf(self, node: Router):
        """!
        @brief Setup LDP and OSPF on router.

        @param node node.
        """
        self._log('Setting up LDP and OSPF on as{}/{}'.format(node.getAsn(), node.getName()))

        node.setPrivileged(True)
        node.addSoftware('frr')

        ospf_ifaces = ''
        ldp_ifaces = ''
        mpls_iface_list = []

        # todo mask network from ospf?
        for iface in node.getInterfaces():
            net = iface.getNet()
            if net.getType() == NetworkType.InternetExchange: continue
            if not (True in (node.getRole() == NodeRole.Router for node in net.getAssociations())): continue
            ospf_ifaces += MplsFileTemplates['frr_config_ospf_iface'].format(interface = net.getName())
            ldp_ifaces += MplsFileTemplates['frr_config_ldp_iface'].format(interface = net.getName())
            mpls_iface_list.append(net.getName())
            net.setMtu(9000)

        node.setFile('/etc/frr/frr.conf', MplsFileTemplates['frr_config'].format(
            loopbackAddress = node.getLoopbackAddress(),
            ospfInterfaces = ospf_ifaces,
            ldpInterfaces = ldp_ifaces
        ))

        node.setFile('/frr_start', MplsFileTemplates['frr_start_script'])
        node.setFile('/mpls_ifaces.txt', '\n'.join(mpls_iface_list))
        node.addStartCommand('chmod +x /frr_start')
        node.addStartCommand('/frr_start')

    def __setUpIbgpMesh(self, nodes: List[Router]):
        """!
        @brief Setup IBGP full mesh.

        @param node list of nodes.
        """

        self._log('Setting up iBGP full mesh for edge nodes...')
        for local in nodes:

            n = 1
            for remote in nodes:
                if local == remote: continue

                local.addTable('t_bgp')
                local.addTablePipe('t_bgp')
                local.addTablePipe('t_direct', 't_bgp')
                local.addProtocol('bgp', 'ibgp{}'.format(n), MplsFileTemplates['bird_ibgp_peer'].format(
                    localAddress = local.getLoopbackAddress(),
                    peerAddress = remote.getLoopbackAddress(),
                    asn = local.getAsn()
                ))

                n += 1

    def onRender(self):
        for asn in self.__enabled:
            if self.__reg.has('seedsim', 'layer', 'Ospf'):
                self._log('Ospf layer exists, masking as{}'.format(asn))
                ospf: Ospf = self.__reg.get('seedsim', 'layer', 'Ospf')
                ospf.maskAsn(asn)

            if self.__reg.has('seedsim', 'layer', 'Ibgp'):
                self._log('Ibgp layer exists, masking as{}'.format(asn))
                ibgp: Ibgp = self.__reg.get('seedsim', 'layer', 'Ibgp')
                ibgp.mask(asn)

            scope = ScopedRegistry(str(asn))
            (enodes, nodes) = self.__getEdgeNodes(scope)

            for n in enodes: self.__setUpLdpOspf(n)
            for n in nodes: self.__setUpLdpOspf(n)
            self.__setUpIbgpMesh(enodes)

    def _doCreateGraphs(self):
        base: Base = self.__reg.get('seedsim', 'layer', 'Base')
        for asn in base.getAsns():
            if asn not in self.__enabled: continue
            asobj = base.getAutonomousSystem(asn)
            asobj.createGraphs()
            l2graph = asobj.getGraph('AS{}: Layer 2 Connections'.format(asn))
            mplsgraph = self._addGraph('AS{}: MPLS Topology'.format(asn), False)
            mplsgraph.copy(l2graph)
            for edge in mplsgraph.edges:
                edge.style = 'dotted'

            scope = ScopedRegistry(str(asn))
            (enodes, _) = self.__getEdgeNodes(scope)
            
            while len(enodes) > 0:
                a = enodes.pop()
                for b in enodes:
                    mplsgraph.addEdge('Router: {}'.format(a.getName()), 'Router: {}'.format(b.getName()), style = 'solid')

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'MplsLayer:\n'
        
        indent += 4
        out += ' ' * indent
        out += 'Enabled on:\n'

        indent += 4
        for asn in self.__enabled:
            out += ' ' * indent
            out += 'as{}\n'.format(asn)

        return out
            


from seedsim.core import Registry
from .Compiler import Compiler
from .DistributedDocker import DistributedDocker
from typing import Dict
from hashlib import md5
from os import mkdir, chdir, chmod

GcpDistributedDockerFileTemplates: Dict[str, str] = {}

GcpDistributedDockerFileTemplates['_tf_scripts/get-swmtkn'] = '''\
#!/bin/bash
host="`jq -cr '.host'`"
token="`ssh -i ssh_key -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null "seed@$host" 'sudo docker swarm join-token worker -q'`"
jq -ncr --arg token "$token" '{"token":$token}'
exit 0
'''

GcpDistributedDockerFileTemplates['_tf_scripts/ssh-keygen'] = '''\
#!/bin/bash
[ ! -e ssh_key ] && ssh-keygen -b 2048 -t rsa -f ./ssh_key -q -N ''
jq -ncr --arg private_key "`cat ssh_key`" --arg public_key "`cat ssh_key.pub`" '{"private_key":$private_key, "public_key":$public_key}'
exit 0
'''

GcpDistributedDockerFileTemplates['variables.tf'] = '''\
variable "project" {
  type = string
  description = "GCP project ID"
}

variable "region" {
  type = string
  description = "GCP region"
}

variable "zone" {
  type = string
  description = "GCP zone"
}

variable "credentials_file" {
  type = string
  description = "Path to the JSON credentials file (https://console.cloud.google.com/apis/credentials/serviceaccountkey)"
}
'''

GcpDistributedDockerFileTemplates['main.tf'] = '''\
terraform {
  required_providers {
    google = {
      source = "hashicorp/google"
    }
  }
}

provider "google" {
  version = "3.5.0"

  credentials = file(var.credentials_file)

  project = var.project
  region  = var.region
  zone    = var.zone
}
'''

GcpDistributedDockerFileTemplates['network.tf'] = '''\
resource "google_compute_network" "swarm" {
  name = "seedsim-swarm"
}

resource "google_compute_firewall" "swarm" {
  name    = "seedsim-swarm-firewall"
  network = google_compute_network.swarm.name

  allow {
    protocol = "all"
  }

  source_ranges = ["0.0.0.0/0"]
}
'''

GcpDistributedDockerFileTemplates['data.tf'] = '''\
data "external" "ssh_keys" {
  program = ["_tf_scripts/ssh-keygen"]
}

data "external" "swarm_tokens" {
  program = ["_tf_scripts/get-swmtkn"]

  query = {
    host = google_compute_instance.manager.network_interface[0].access_config[0].nat_ip
  }
}
'''

GcpDistributedDockerFileTemplates['manager_tf_template'] = '''\
resource "google_compute_instance" "manager" {{
  name = "manager"
  machine_type = "{machineType}"

  boot_disk {{
    initialize_params {{
      image = "debian-cloud/debian-10"
      size  = 16
    }}
  }}

  metadata = {{
    ssh-keys = "seed:${{data.external.ssh_keys.result.public_key}}"
  }}

  network_interface {{
    network = google_compute_network.swarm.name
    access_config {{
    }}
  }}

  connection {{
    host = self.network_interface[0].access_config[0].nat_ip
    type = "ssh"
    user = "seed"
    private_key = data.external.ssh_keys.result.private_key
  }}

  provisioner "file" {{
      source = "_containers/ix"
      destination = "/tmp/"
  }}

  provisioner "remote-exec" {{
    inline = [
      "sudo apt-get update",
      "sudo apt-get -qq --no-install-recommends install apt-transport-https ca-certificates curl gnupg-agent software-properties-common",
      "curl -fsSL https://download.docker.com/linux/debian/gpg | sudo apt-key add -",
      "sudo add-apt-repository \\"deb [arch=amd64] https://download.docker.com/linux/debian `lsb_release -cs` stable\\"",
      "sudo apt-get update",
      "sudo apt-get -qq --no-install-recommends install docker-ce docker-ce-cli containerd.io docker-compose",
      "sudo docker swarm init",
      "sudo modprobe mpls_router",
      "cd /tmp/ix",
      "sudo docker-compose up -d"
    ]
  }}

  depends_on = [google_compute_firewall.swarm]
}}
'''

GcpDistributedDockerFileTemplates['worker_tf_template'] = '''\
resource "google_compute_instance" "worker-as{name}" {{
  name = "worker-as{name}"
  machine_type = "{machineType}"

  boot_disk {{
    initialize_params {{
      image = "debian-cloud/debian-10"
      size  = 16
    }}
  }}

  metadata = {{
    ssh-keys = "seed:${{data.external.ssh_keys.result.public_key}}"
  }}

  network_interface {{
    network = google_compute_network.swarm.name
    access_config {{
    }}
  }}

  connection {{
    host = self.network_interface[0].access_config[0].nat_ip
    type = "ssh"
    user = "seed"
    private_key = data.external.ssh_keys.result.private_key
  }}

  provisioner "file" {{
      source = "_containers/{name}"
      destination = "/tmp/"
  }}

  provisioner "remote-exec" {{
    inline = [
      "sudo apt-get update",
      "sudo apt-get -qq --no-install-recommends install apt-transport-https ca-certificates curl gnupg-agent software-properties-common",
      "curl -fsSL https://download.docker.com/linux/debian/gpg | sudo apt-key add -",
      "sudo add-apt-repository \\"deb [arch=amd64] https://download.docker.com/linux/debian `lsb_release -cs` stable\\"",
      "sudo apt-get update",
      "sudo apt-get -qq --no-install-recommends install docker-ce docker-ce-cli containerd.io docker-compose",
      "sudo docker swarm join --token ${{data.external.swarm_tokens.result.token}} ${{google_compute_instance.manager.network_interface[0].network_ip}}:2377",
      "sudo modprobe mpls_router",
      "cd /tmp/{name}",
      "sudo docker-compose up -d"
    ]
  }}
}}
'''

class GcpDistributedDocker(Compiler):
    """!
    @brief The GcpDistributedDocker compiler class.

    GcpDistributedDocker is one of the compiler driver. It compiles the lab to
    sets of docker containers, and generate Terraform configuration for
    deploying the lab to GCP.
    """

    def getName(self) -> str:
        return 'GcpDistributedDocker'

    def __init_tf(self):
        """!
        @brief Get files required by Terraform ready.
        """
        self._log('initializing terraform environment...')
        mkdir('_tf_scripts')
        for file in ['_tf_scripts/get-swmtkn', '_tf_scripts/ssh-keygen', 'variables.tf', 'main.tf', 'network.tf', 'data.tf']:
            print(GcpDistributedDockerFileTemplates[file], file=open(file, 'w'))
        
        for exfile in ['_tf_scripts/get-swmtkn', '_tf_scripts/ssh-keygen']:
            chmod(exfile, 0o755)

    def __make_tf(self):
        """!
        @brief Generate TF config for docker hosts.
        """
        self._log('generating terraform configurations...')

        print(GcpDistributedDockerFileTemplates['manager_tf_template'].format(
            machineType = "f1-micro" # todo
        ), file=open('manager.tf', 'w'))

        scopes = set()

        for (scope, type, _) in Registry().getAll().keys():
            if scope == 'ix': continue
            if type == 'net' or type == 'hnode' or type == 'rnode' or type == 'snode':
                scopes.add(scope)

        for scope in scopes:
            print(GcpDistributedDockerFileTemplates['worker_tf_template'].format(
            machineType = "f1-micro", # todo
            name = scope
        ), file=open('worker-as{}.tf'.format(scope), 'w'))

    def _doCompile(self):
        dcomp = DistributedDocker()
        self.__init_tf()
        self._log('generating container configurations...')
        dcomp.compile('_containers')
        self.__make_tf()


from __future__ import annotations
from .Service import Server, Service
from typing import List

class RpkiServer(Server):
    """!
    @brief RPKI server.
    """

    def newChild(self, handle: str, delegations: List[str]) -> RpkiServer:
        """!
        @brief Create a new child.

        @param handle Child handle.
        @param delegations IPv4 prefixes to delegate to the child.

        @returns child server.

        @throws AssertionError if child handle exists.
        """

    def hadChild(self, handle: str) -> bool:
        """!
        @brief Tesi if child exists.

        @param handle Child handle.

        @returns True if exist, False otherwise.
        """

    def getChild(self, handle: str) -> RpkiServer:
        """!
        @brief Get a child.

        @param handle Child handle.

        @returns child server.
        
        @throws AssertionError if child handle does not exist.
        """
        pass

    def addRoa(self, prefix: str, asn: int, maxLength: int):
        """!
        @brief Add new ROA (route origin authorization).

        @param prefix prefix.
        @param asn asn.
        @param maxLength max prefix length.
        """
        pass

    def hostTa(self):
        """!
        @brief Host TA (trust anchor) on this node.
        """
        pass

class RpkiService(Service):
    """!
    @brief RPKI service.
    """




from .Layer import Layer
from .DomainNameService import DomainNameService
from typing import Set, Dict, List
from seedsim.core import ScopedRegistry, Node

DnssecFileTemplates: Dict[str, str] = {}

DnssecFileTemplates['enable_dnssec_script'] = """\
#!/bin/bash
rm -fr /etc/bind/keys 
mkdir /etc/bind/keys
cd /etc/bind/keys
rndc freeze
while read -r zonename; do {
    [ -z "$zonename" ] && continue
    zonefile="$zonename"
    [ "$zonename" = "." ] && zonefile="root"
    echo "setting up DNSSEC for "$zonename"..."
    sed -i 's|zone "'"$zonename"'" { type master; file "/etc/bind/zones/'"$zonefile"'"; allow-update { any; }; };|zone "'"$zonename"'" { type master; file "/etc/bind/zones/'"$zonefile"'"; allow-update { any; }; key-directory "/etc/bind/keys"; auto-dnssec maintain; inline-signing yes; };|' /etc/bind/named.conf.zones
    dnssec-keygen -a NSEC3RSASHA1 -b 2048 -n ZONE "$zonename"
    dnssec-keygen -f KSK -a NSEC3RSASHA1 -b 4096 -n ZONE "$zonename"
}; done < /dnssec_zones.txt

chown -R bind:bind /etc/bind/keys
rndc thaw
rndc reload

while read -r zonename; do {
    [ -z "$zonename" ] && continue
    [ "$zonename" = "." ] && continue
    pzonename="`tr '.' '\\n' <<< "$zonename" | sed '1d' | tr '\\n' '.' | sed -e 's/\\.\\.$/./'`"
    while true; do {
        pns="`dig +short NS "$pzonename"`" || pns=''
        [ -z "$pns" ] && echo "cannot get NS for parent zone ($pzonename), retrying in 1 second..." || break
        sleep 1
    }; done
    dig +short NS "$pzonename" | while read -r ns; do dig +short "$ns"; done | while read -r nsaddr; do {
        dss="`dig @127.0.0.1 dnskey "$zonename" | dnssec-dsfromkey -f- "$zonename" | sed 's/IN/300/; s/^/update add /;'`"
        echo "$dss"
        echo "submiting DS record to parent zone $nsaddr..."
        while true; do {
            cat << UPDATE | nsupdate && echo "parent accepted the update." && break
server $nsaddr
zone $pzonename
$dss
send
UPDATE
            echo "submission failed, retrying in 1 second..."
            sleep 1
        }; done
    };done
}; done < /dnssec_zones.txt
"""

class Dnssec(Layer):
    """!
    @brief The Dnssec (DNSSEC) layer.

    This layer helps setting up DNSSEC. It works by signing the zones and send
    the DS record to parent(s) with nsupdate. Note that to build a DNSSEC
    infrastructure, you will need to sign the entire chain. 
    """

    __zonenames: Set[str]
    __reg = ScopedRegistry('seedsim')

    def __init__(self):
        """!
        @brief Dnssec layer constructor.
        """
        self.__zonenames = set()
        self.addDependency('DomainNameService', False, False)

    def getName(self):
        return 'Dnssec'
    
    def enableOn(self, zonename: str):
        """!
        @brief Enable DNSSEC on the given zone.

        @param zonename zonename. 
        """
        if zonename[-1] != '.': zonename += '.'
        self.__zonenames.add(zonename)

    def onRender(self):
        dns: DomainNameService = self.__reg.get('layer', 'DomainNameService')
        nodes: Set[Node] = set()
        for zonename in self.__zonenames:
            self._log('Looking for server hosting "{}"...'.format(zonename))
            server = dns.getServerByZoneName(zonename)
            
            assert server != None, 'no server found for dnssec-enabled zone {}'.format(zonename)
            node = server.getNode()
            (scope, _, name) = node.getRegistryInfo()
            self._log('Setting up DNSSEC for "{}" on as{}/{}'.format(zonename, scope, name))
            nodes.add(node)
            node.appendFile('/dnssec_zones.txt', '{}\n'.format(zonename))

        for node in nodes:
            node.appendFile('/enable_dnssec', DnssecFileTemplates['enable_dnssec_script'])
            node.addStartCommand('chmod +x /enable_dnssec')
            node.addStartCommand('/enable_dnssec')

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'DnssecLayer:\n'

        indent += 4

        out += ' ' * indent
        out += 'DNSSEC-enabled zones:\n'

        for zonename in self.__zonenames:
            out += ' ' * indent
            out += '{}\n'.format(zonename)

        return out
            




from .Layer import Layer
from .Base import Base
from .Routing import Routing, Router
from .Ebgp import Ebgp
from .Ospf import Ospf
from .Ibgp import Ibgp
from .Service import Service, Server
from .WebService import WebService, WebServer
from .DomainNameService import DomainNameServer, DomainNameService, Zone
from .DomainNameCachingService import DomainNameCachingServer, DomainNameCachingService
from .Reality import Reality
from .CymruIpOrigin import CyrmuIpOriginService, CyrmuIpOriginServer
from .Dnssec import Dnssec
from .ReverseDomainNameService import ReverseDomainNameService, ReverseDomainNameServer
from .Mpls import Mpls


from .Printable import Printable
from .enums import NetworkType, NodeRole
from typing import Generator

class AddressAssignmentConstraint(Printable):
    """!
    AddressAssignmentConstraint class.

    This class defines how IP addresses should be assign to network interfaces.
    Derive from this class to change the default behavior.
    """

    def __range(self, a: int, b: int, s: int = 1):
        for n in range(a, b, s): yield n

    def __always(self, a: int):
        while True: yield a

    def getOffsetGenerator(self, type: NodeRole) -> Generator[int, None, None]:
        """!
        @brief Get IP offset generator for a type of node.

        @todo Handle pure-internal routers.

        @param type type of the node.
        @returns An int generator that generates IP address offset.
        @throws ValueError if try to get generator of IX interface.
        """

        if type == NodeRole.Host: return self.__range(71, 99)
        if type == NodeRole.Router: return self.__range(254, 200, -1)

        raise ValueError("IX IP assigment must done with mapIxAddress().")

    def mapIxAddress(self, asn: int) -> int:
        """!
        @brief Map ASN to IP address in IX peering LAN.

        @param asn ASN of IX participant.
        @returns offset.
        @throws AssertionError if can't map ASN to IP address.
        """
        assert asn >= 2 and asn <= 254, "can't map ASN {} to IX address.".format(asn)
        return asn

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'AddressAssignmentConstraint: Default Constraint\n'

        return out


from seedsim.core import Registry, ScopedRegistry, Node, Network
from .Compiler import Compiler
from typing import Dict
from hashlib import md5
from os import mkdir, chdir, rmdir

DistributedDockerCompilerFileTemplates: Dict[str, str] = {}

DistributedDockerCompilerFileTemplates['dockerfile'] = """\
FROM ubuntu:20.04
ARG DEBIAN_FRONTEND=noninteractive
RUN apt-get update
RUN apt-get install -y zsh curl
RUN curl -L https://grml.org/zsh/zshrc > /root/.zshrc
RUN echo 'exec zsh' > /root/.bashrc
"""

DistributedDockerCompilerFileTemplates['start_script'] = """\
#!/bin/bash
{startCommands}
echo "ready! run 'docker exec -it $HOSTNAME /bin/zsh' to attach to this node" >&2
tail -f /dev/null
"""

DistributedDockerCompilerFileTemplates['compose'] = """\
version: "3.4"
services:
{services}
networks:
{networks}
"""

DistributedDockerCompilerFileTemplates['compose_service'] = """\
    {nodeId}:
        build: ./{nodeId}
        cap_add:
            - ALL
        sysctls:
            - net.ipv4.ip_forward=1
        privileged: {privileged}
        networks:
{networks}{ports}
"""
DistributedDockerCompilerFileTemplates['compose_ports'] = """\
        ports:
{portList}
"""

DistributedDockerCompilerFileTemplates['compose_port'] = """\
            - {hostPort}:{nodePort}/{proto}
"""

DistributedDockerCompilerFileTemplates['compose_service_network'] = """\
            {netId}:
                ipv4_address: {address}
"""

DistributedDockerCompilerFileTemplates['compose_network'] = """\
    {netId}:
        driver_opts:
            com.docker.network.driver.mtu: {mtu}
        ipam:
            config:
                - subnet: {prefix}
"""

DistributedDockerCompilerFileTemplates['compose_network_ix_worker'] = """\
    {netId}:
        external:
            name: sim_ix_{netId}
        driver: overlay
"""

DistributedDockerCompilerFileTemplates['compose_network_ix_master'] = """\
    {netId}:
        driver: overlay
        ipam:
            config:
                - subnet: {prefix}
"""

class DistributedDocker(Compiler):
    """!
    @brief The DistributedDocker compiler class.

    DistributedDocker is one of the compiler driver. It compiles the lab to
    docker containers. This compiler will generate one set of containers with
    their docker-compose.yml for each AS, enable you to run the simulator
    distributed. 

    This works by making every IX network overlay network. 
    """

    def __init__(self):
        """!
        @brief DistributedDocker compiler constructor.
        """

    def getName(self) -> str:
        return "DistributedDocker"

    def __contextToPrefix(self, scope: str, type: str) -> str:
        """!
        @brief Convert context to prefix.

        @param scope scope.
        @param type type.

        @returns prefix string.
        """
        return '{}_{}_'.format(type, scope)

    def __addFile(self, path: str, content: str) -> str:
        """!
        @brief Stage file to local folder and return Dockerfile command.

        @param path path to file. (in container)
        @param content content of the file.
        """

        staged_path = md5(path.encode('utf-8')).hexdigest()
        print(content, file=open(staged_path, 'w'))
        return 'COPY {} {}\n'.format(staged_path, path)

    def __compileNode(self, node: Node) -> str:
        """!
        @brief Compile a single node.

        @param node node to compile.
        """
        (scope, type, _) = node.getRegistryInfo()
        prefix = self.__contextToPrefix(scope, type)
        real_nodename = '{}{}'.format(prefix, node.getName())

        node_nets = ''

        for iface in node.getInterfaces():
            net = iface.getNet()
            (netscope, _, _) = net.getRegistryInfo()
            net_prefix = self.__contextToPrefix(netscope, 'net')
            real_netname = '{}{}'.format(net_prefix, net.getName())

            node_nets += DistributedDockerCompilerFileTemplates['compose_service_network'].format(
                netId = real_netname,
                address = iface.getAddress()
            )
        
        _ports = node.getPorts()
        ports = ''
        if len(_ports) > 0:
            lst = ''
            for (h, n, p) in _ports:
                lst += DistributedDockerCompilerFileTemplates['compose_port'].format(
                    hostPort = h,
                    nodePort = n,
                    proto = p
                )
            ports = DistributedDockerCompilerFileTemplates['compose_ports'].format(
                portList = lst
            )
        
        dockerfile = DistributedDockerCompilerFileTemplates['dockerfile']
        mkdir(real_nodename)
        chdir(real_nodename)

        commsoft = node.getCommonSoftware()
        if len(commsoft) > 0: dockerfile += 'RUN apt-get install -y --no-install-recommends {}\n'.format(' '.join(sorted(commsoft)))

        soft = node.getSoftwares()
        if len(soft) > 0: dockerfile += 'RUN apt-get install -y --no-install-recommends {}\n'.format(' '.join(sorted(soft)))

        for cmd in node.getBuildCommands(): dockerfile += 'RUN {}\n'.format(cmd)

        start_commands = ''
        for (cmd, fork) in node.getStartCommands():
            start_commands += '{}{}\n'.format(cmd, ' &' if fork else '')

        dockerfile += self.__addFile('/start.sh', DistributedDockerCompilerFileTemplates['start_script'].format(
            startCommands = start_commands
        ))

        dockerfile += 'RUN chmod +x /start.sh\n'

        for file in node.getFiles():
            (path, content) = file.get()
            dockerfile += self.__addFile(path, content)

        dockerfile += 'CMD ["/start.sh"]\n'
        print(dockerfile, file=open('Dockerfile', 'w'))

        chdir('..')

        return DistributedDockerCompilerFileTemplates['compose_service'].format(
            nodeId = real_nodename,
            networks = node_nets,
            privileged = 'true' if node.isPrivileged() else 'false',
            ports = ports
        )

    def __compileNet(self, net: Network) -> str:
        (scope, _, _) = net.getRegistryInfo()
        return DistributedDockerCompilerFileTemplates['compose_network'].format(
            netId = '{}{}'.format(self.__contextToPrefix(scope, 'net'), net.getName()),
            prefix = net.getPrefix(),
            mtu = net.getMtu()
        )

    def __compileIxNetMaster(self, net) -> str:
        (scope, _, _) = net.getRegistryInfo()
        return DistributedDockerCompilerFileTemplates['compose_network_ix_master'].format(
            netId = '{}{}'.format(self.__contextToPrefix(scope, 'net'), net.getName()),
            prefix = net.getPrefix()
        )

    def __compileIxNetWorker(self, net) -> str:
        (scope, _, _) = net.getRegistryInfo()
        return DistributedDockerCompilerFileTemplates['compose_network_ix_worker'].format(
            netId = '{}{}'.format(self.__contextToPrefix(scope, 'net'), net.getName())
        )

    def _doCompile(self):
        registry = Registry()
        scopes = set()
        for (scope, _, _) in registry.getAll().keys(): scopes.add(scope)

        ix_nets = ''

        for ixnet in ScopedRegistry('ix').getByType('net'):
            ix_nets += self.__compileIxNetWorker(ixnet)

        for scope in scopes:
            mkdir(scope)
            chdir(scope)

            services = ''
            networks = ''

            for ((_scope, type, name), obj) in registry.getAll().items():
                if _scope != scope: continue

                if type == 'rnode':
                    self._log('compiling router node {} for as{}...'.format(name, scope))
                    services += self.__compileNode(obj)

                if type == 'hnode':
                    self._log('compiling host node {} for as{}...'.format(name, scope))
                    services += self.__compileNode(obj)

                if type == 'rs':
                    self._log('compiling rs node for {}...'.format(name))
                    services += self.__compileNode(obj)

                if type == 'snode':
                    self._log('compiling service node {}...'.format(name))
                    services += self.__compileNode(obj)

                if type == 'net':
                    self._log('creating network: {}/{}...'.format(scope, name))
                    networks += self.__compileIxNetMaster(obj) if scope == 'ix' else self.__compileNet(obj)

            if len(services) > 0 or len(networks) > 0 :
                if scope != 'ix': networks += ix_nets
                self._log('creating docker-compose.yml...'.format(scope, name))
                print(DistributedDockerCompilerFileTemplates['compose'].format(
                    services = services,
                    networks = networks
                ), file=open('docker-compose.yml', 'w'))

                print('COMPOSE_PROJECT_NAME=sim_{}'.format(scope), file=open('.env', 'w'))

            chdir('..')

            if services == '' and networks == '': rmdir(scope)



from seedsim.core import Node, ScopedRegistry
from .Service import Service, Server
from .DomainNameService import DomainNameService
from typing import List, Dict

DomainNameCachingServiceFileTemplates: Dict[str, str] = {}

DomainNameCachingServiceFileTemplates['named_options'] = '''\
options {
	directory "/var/cache/bind";
	recursion yes;
	dnssec-validation no;
    empty-zones-enable no;
	allow-query { any; };
};
'''

class DomainNameCachingServer(Server):
    """!
    @brief Caching DNS server (i.e., Local DNS server)

    @todo DNSSEC
    """

    __root_servers: List[str]
    __node: Node

    def __init__(self, node: Node):
        """!
        @brief DomainNameCachingServer constructor.

        @param node node to install on.
        """
        self.__root_servers = []
        self.__node = node

    def setRootServers(self, servers: List[str]):
        """!
        @brief Change root server hint.

        By defualt, the caching server uses the root hint file shipped with
        bind9. Use this method to override root hint. Note that if autoRoot is
        set to true in DomainNameCachingService, manual changes will be
        overrided.

        @param servers list of IP addresses of the root servers.
        """
        self.__root_servers = servers

    def getRootServers(self) -> List[str]:
        """!
        @brief Get root server list.

        By defualt, the caching server uses the root hint file shipped with
        bind9. Use setRootServers to override root hint.

        This method will return list of servers set by setRootServers, or an
        empty list if not set.
        """
        return self.__root_servers

    def install(self, setResolvconf: bool):
        """!
        @brief Handle installation.

        @param setResolvconf set resolv.conf of all other nodes in the AS.

        @throw AssertionError if setResolvconf is true but current node has no
        IP address.
        """
        self.__node.addSoftware('bind9')
        self.__node.setFile('/etc/bind/named.conf.options', DomainNameCachingServiceFileTemplates['named_options'])
        if len(self.__root_servers) > 0:
            hint = '\n'.join(self.__root_servers)
            self.__node.setFile('/usr/share/dns/root.hints', hint)
            self.__node.setFile('/etc/bind/db.root', hint)
        self.__node.addStartCommand('service named start')
        if not setResolvconf: return
        (scope, _, _) = self.__node.getRegistryInfo()
        sr = ScopedRegistry(scope)
        ifaces = self.__node.getInterfaces()
        assert len(ifaces) > 0, 'Node {} has no IP address.'.format(self.__node.getName())
        addr = ifaces[0].getAddress()

        for rnode in sr.getByType('rnode'):
            rnode.appendFile('/etc/resolv.conf.new', 'nameserver {}\n'.format(addr))
            if 'cp /etc/resolv.conf.new /etc/resolv.conf' not in rnode.getStartCommands():
                rnode.addStartCommand('cp /etc/resolv.conf.new /etc/resolv.conf')

        for hnode in sr.getByType('hnode'):
            if 'cp /etc/resolv.conf.new /etc/resolv.conf' not in hnode.getStartCommands():
                hnode.addStartCommand('cp /etc/resolv.conf.new /etc/resolv.conf')
            hnode.appendFile('/etc/resolv.conf.new', 'nameserver {}\n'.format(addr))

class DomainNameCachingService(Service):
    """!
    @brief Caching DNS (i.e., Local DNS)

    @todo DNSSEC
    """

    __auto_root: bool
    __set_resolvconf: bool
    __reg = ScopedRegistry('seedsim')
    __servers: List[DomainNameCachingServer]

    def __init__(self, autoRoot: bool = True, setResolvconf: bool = False):
        """!
        @brief DomainNameCachingService constructor.

        @param autoRoot (optional) find root zone name servers automaically.
        True by defualt, if true, DomainNameCachingService will find root NS in
        DomainNameService and use them as root.
        @param setResolvconf (optional) set all nodes in the AS to use local DNS
        node in the AS by overrideing resolv.conf. Default to false.
        """

        self.__auto_root = autoRoot
        self.__set_resolvconf = setResolvconf
        self.__servers = []
        self.addDependency('Base', False, False)
        if autoRoot:
            self.addDependency('DomainNameService', False, False)

    def getName(self) -> str:
        return 'DomainNameCachingService'

    def getConflicts(self) -> List[str]:
        return ['DomainNameService']

    def _doInstall(self, node: Node) -> DomainNameCachingServer:
        """!
        @brief Install the cache service on given node.

        @param node node to install the cache service on.

        @returns Handler of the installed cache service.
        @throws AssertionError if node is not host node.
        """
        server: DomainNameCachingServer = node.getAttribute('__domain_name_cacheing_service_server')
        if server != None: return server
        server = DomainNameCachingServer(node)
        self.__servers.append(server)
        node.setAttribute('__domain_name_cacheing_service_server', server)
        return server

    def onRender(self):
        root_servers: List[str] = []
        if self.__auto_root:
            dns_layer: DomainNameService = self.__reg.get('layer', 'DomainNameService')
            root_zone = dns_layer.getRootZone()
            root_servers = root_zone.getGuleRecords()
        
        for server in self.__servers:
            server.setRootServers(root_servers)
            server.install(self.__set_resolvconf)

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'DomainNameCachingService:\n'

        indent += 4

        out += ' ' * indent
        out += 'Configure root hint: {}\n'.format(self.__auto_root)

        out += ' ' * indent
        out += 'Set resolv.conf: {}\n'.format(self.__set_resolvconf)

        return out


import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name = "seedsim",
    version = "0.0.5",
    author = "Honghao Zeng",
    author_email = "hozeng@syr.edu",
    description = "SEED Internet Simulator",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/seed-labs/seed-simulator",
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        'requests'
    ],
    python_requires = '>=3.6'
)



from .AddressAssignmentConstraint import AddressAssignmentConstraint
from .AutonomousSystem import AutonomousSystem
from .InternetExchange import InternetExchange
from .Network import Network
from .Node import Node, File, Interface
from .Printable import Printable
from .Registry import Registry, ScopedRegistry, Registrable
from .Graphable import Graphable, Graph, Vertex, Edge


from .Service import Service, Server
from .DomainNameService import DomainNameService, DomainNameServer, Zone
from seedsim.core import ScopedRegistry, Registry, Node
from typing import List

class ReverseDomainNameServer(Server):
    """!
    @brief Reverse DNS server.
    """

    __node: Node

    def __init__(self, node: Node):
        """!
        @brief ReverseDnsServer constructor.

        @param node node to install on.
        """
        self.__node = node

    def getNode(self) -> Node:
        """!
        @brief get node.

        @returns node.
        """
        return self.__node

class ReverseDomainNameService(Service):
    """!
    @brief Reverse DNS. This service hosts the in-addr.arpa. zone and resolve
    IP addresses to nodename-netname.nodetype.asn.net
    """

    __servers: List[ReverseDomainNameServer]
    __reg = ScopedRegistry('seedsim')

    def __init__(self):
        """!
        @brief ReverseDomainNameService constructor
        """
        self.__records = []
        self.__servers = []
        self.addDependency('DomainNameService', True, False)
        self.addDependency('Base', False, False)

    def getName(self) -> str:
        return 'ReverseDomainNameService'

    def _doInstall(self, node: Node) -> ReverseDomainNameServer:
        server = ReverseDomainNameServer(node)
        self.__servers.append(server)

    def onRender(self):
        self._log('Creating "in-addr.arpa." zone...')
        dns: DomainNameService = self.__reg.get('layer', 'DomainNameService')
        zone = dns.getZone('in-addr.arpa.')

        self._log('Collecting IP addresses...')
        for ([scope, type, name], obj) in Registry().getAll().items():
            if type != 'rnode' and type != 'hnode': continue
            self._log('Collecting {}/{}/{}...'.format(scope, type, name))

            if scope == 'ix':
                scope = name
                name = 'rs'
            else: scope = 'as' + scope

            node: Node = obj
            for iface in node.getInterfaces():
                addr = '.'.join(reversed(str(iface.getAddress()).split('.')))
                netname = iface.getNet().getName()
                record = '{} PTR {}-{}.{}.{}.net.'.format(addr, name, netname, type, scope)
                zone.addRecord(record)

        self._log('Setting up "in-addr.arpa." server nodes...')
        for server in self.__servers:
            dns.hostZoneOn('in-addr.arpa.', server.getNode())

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'ReverseDomainNameService:\n'
        
        indent += 4
        out += ' ' * indent
        out += 'Installed on:\n'

        indent += 4
        for server in self.__servers:
            (asn, _, name) = server.getNode().getRegistryInfo()
            out += ' ' * indent
            out += 'as{}/{}\n'.format(asn, name)

        return out


        
    


from .Printable import Printable
from .Network import Network
from .enums import NodeRole, NetworkType
from .Registry import Registry, ScopedRegistry, Registrable
from ipaddress import IPv4Address
from typing import List, Dict, Set, Tuple

DEFAULT_SOFTWARES: List[str] = ['curl', 'nano', 'vim-nox', 'mtr-tiny', 'iproute2', 'iputils-ping', 'tcpdump', 'termshark', 'dnsutils', 'jq', 'ipcalc']

class File(Printable):
    """!
    @brief File class.

    This class represents a file on a node.
    """

    __content: str
    __path: str

    def __init__(self, path: str, content: str = ''):
        """!
        @brief File constructor.

        Put a file onto a node.

        @param path path of the file.
        @param content content of the file.
        """
        self.__path = path
        self.__content = content

    def setPath(self, path: str):
        """!
        @brief Update file path.

        @param path new path.
        """
        self.__path = path

    def setContent(self, content: str):
        """!
        @brief Update file content.

        @param content content.
        """
        self.__content = content

    def appendContent(self, content: str):
        """!
        @brief Append to file.

        @param content content.
        """
        self.__content += content

    def get(self) -> (str, str):
        """!
        @brief Get file path and content.

        @returns a tuple where the first element is path and second element is 
        content
        """
        return (self.__path, self.__content)

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += "{}:\n".format(self.__path)
        indent += 4
        for line in self.__content.splitlines():
            out += ' ' * indent
            out += '> '
            out += line
            out += '\n'
        return out

class Interface(Printable):
    """!
    @brief Interface class.

    This class represents a network interface card.
    """

    __network: Network
    __address: IPv4Address

    __latency: int       # in ms
    __bandwidth: int     # in bps
    __drop: float        # percentage

    def __init__(self, net: Network):
        """!
        @brief Interface constructor.

        @param net network to connect to.
        """
        self.__address = None
        self.__network = net
        (l, b, d) = net.getDefaultLinkProperties()
        self.__latency = l
        self.__bandwidth = b
        self.__drop = d

    def setLinkProperties(self, latency: int = 0, bandwidth: int = 0, packetDrop: float = 0):
        """!
        @brief Set link properties.

        @param latency (optional) latency to add to the link in ms, default 0.
        @param bandwidth (optional) egress bandwidth of the link in bps, 0 for unlimited, default 0.
        @param packetDrop (optional) link packet drop as percentage, 0 for unlimited, default 0.
        """

        assert latency >= 0, 'invalid latency'
        assert bandwidth >= 0, 'invalid bandwidth'
        assert packetDrop >= 0 and packetDrop <= 100, 'invalid packet drop'

        self.__latency = latency
        self.__bandwidth = bandwidth
        self.__drop = packetDrop

    def getLinkProperties(self) -> Tuple[int, int, int]:
        """!
        @brief Get link properties.

        @returns tuple (latency, bandwidth, packet drop)
        """
        return (self.__latency, self.__bandwidth, self.__drop)

    def getNet(self) -> Network:
        """!
        @brief Get the network that this interface attached to.

        @returns network.
        """
        return self.__network

    def setAddress(self, address: IPv4Address):
        """!
        @brief Set IPv4 address of this interface.

        @param address address.
        """
        self.__address = address

    def getAddress(self):
        """!
        @brief Get IPv4 address of this interface.

        @returns address.
        """
        return self.__address

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'Interface:\n'

        indent += 4
        out += ' ' * indent
        out += 'Connected to: {}\n'.format(self.__network.getName())
        out += ' ' * indent
        out += 'Address: {}\n'.format(self.__address)
        out += ' ' * indent
        out += 'Link Properties: {}\n'.format(self.__address)

        indent += 4
        out += ' ' * indent
        out += 'Added Latency: {} ms\n'.format(self.__latency)
        out += ' ' * indent
        out += 'Egress Bandwidth Limit: {} bps\n'.format('unlimited' if self.__bandwidth <= 0 else self.__bandwidth)

        return out

class Node(Printable, Registrable):
    """!
    @brief Node base class.

    This class represents a generic node.
    """

    __name: str
    __asn: int
    __role: NodeRole
    __reg: ScopedRegistry
    __greg = Registry()
    __interfaces: List[Interface]
    __files: Dict[str, File]
    __softwares: Set[str]
    __build_commands: List[str]
    __start_commands: List[Tuple[str, bool]]
    __ports: List[Tuple[int, int, str]]
    __privileged: bool
    __common_software: Set[str] = set()

    def __init__(self, name: str, role: NodeRole, asn: int, scope: str = None):
        """!
        @brief Node constructor.

        @name name name of this node.
        @param role role of this node.
        @param asn network that this node belongs to.
        @param scope scope of the node, if not asn.
        """
        self.__interfaces = []
        self.__files = {}
        self.__asn = asn
        self.__reg = ScopedRegistry(scope if scope != None else str(asn))
        self.__role = role
        self.__name = name
        self.__softwares = set()
        self.__build_commands = []
        self.__start_commands = []
        self.__ports = []
        self.__privileged = False

        for soft in DEFAULT_SOFTWARES:
            self.__common_software.add(soft)

    def addPort(self, host: int, node: int, proto: str = 'tcp'):
        """!
        @brief Add port forwarding.

        @param host port of the host.
        @param node port of the node.
        @param proto protocol.
        """
        self.__ports.append((host, node, proto))

    def getPorts(self) -> List[Tuple[int, int]]:
        """!
        @brief Get port forwardings.

        @returns list of tuple of ports (host, node).
        """
        return self.__ports

    def setPrivileged(self, privileged: bool):
        """!
        @brief Set or unset the node as privileged node.

        Some backend like Docker will require the container to be privileged
        in order to do some privileged operations.

        @param privileged (optional) set if node is privileged.
        """
        self.__privileged = privileged

    def isPrivileged(self,) -> bool:
        """!
        @brief Test if node is set to privileged.

        @returns True if privileged, False otherwise.
        """
        return self.__privileged

    def joinNetwork(self, net: Network, address: str = "auto") -> Interface:
        """!
        @brief Connect the node to a network.
        @param net network to connect.
        @param address (optional) override address assigment.

        @throws AssertionError if network does not exist.
        """
        
        if address == "auto": _addr = net.assign(self.__role, self.__asn)
        else: _addr = IPv4Address(address)

        _iface = Interface(net)
        _iface.setAddress(_addr)

        self.__interfaces.append(_iface)
        
        net.associate(self)
        return _iface

    def joinNetworkByName(self, netname: str, address: str = "auto") -> Interface:
        """!
        @brief Connect the node to a network.
        @param netname name of the network.
        @param address (optional) override address assigment.

        @throws AssertionError if network does not exist.
        """
        if self.__reg.has("net", netname):
            return self.joinNetwork(self.__reg.get("net", netname), address)

        if self.__greg.has("ix", "net", netname):
            return self.joinNetwork(self.__greg.get("ix", "net", netname), address)
        
        assert False, 'No such network: {}'.format(netname)

    def getName(self) -> str:
        """!
        @brief Get node name.

        @returns name.
        """
        return self.__name

    def getAsn(self) -> int:
        """!
        @brief Get node parent AS ASN.

        @returns asm.
        """
        return self.__asn

    def getRole(self) -> NodeRole:
        """!
        @brief Get role of current node.

        Get role type of current node. 

        @returns role.
        """
        return self.__role

    def getFile(self, path: str) -> File:
        """!
        @brief Get a file object, and create if not exist.

        @param path file path.
        @returns file.
        """
        if path in self.__files: return self.__files[path]
        self.__files[path] = File(path)
        return self.__files[path]

    def getFiles(self) -> List[File]:
        """!
        @brief Get all files.

        @return list of files.
        """
        return self.__files.values()

    def setFile(self, path: str, content: str):
        """!
        @brief Set content of the file.

        @param path path of the file. Will be created if not exist, and will be
        override if already exist.
        @param content file content.
        """
        self.getFile(path).setContent(content)

    def appendFile(self, path: str, content: str):
        """!
        @brief Append content to a file.

        @param path path of the file. Will be created if not exist.
        @param content content to append.
        """
        self.getFile(path).appendContent(content)

    def addSoftware(self, name: str):
        """!
        @brief Add new software to node.

        @param name software package name.

        Use this to add software to the node. For example, if using the "docker"
        compiler, this will be added as an "apt-get install" line in Dockerfile.
        """
        self.__softwares.add(name)

    def addCommonSoftware(self, name: str):
        """!
        @brief Add new software to all node.

        @param name software package name.

        Use this to add software to all node. For example, if using the "docker"
        compiler, this will be added as an "apt-get install" line in Dockerfile
        of all nodes.
        """
        self.__common_software.add(name)

    def getSoftwares(self) -> Set[str]:
        """!
        @brief Get set of software.

        @returns set of softwares.
        """
        return self.__softwares

    def getCommonSoftware(self) -> Set[str]:
        """!
        @brief Get set of common software.

        @returns set of softwares.
        """
        return self.__common_software

    def addBuildCommand(self, cmd: str):
        """!
        @brief Add new command to build step.

        @param cmd command to add.

        Use this to add build steps to the node. For example, if using the
        "docker" compiler, this will be added as a "RUN" line in Dockerfile.
        """
        self.__build_commands.append(cmd)
    
    def getBuildCommands(self) -> List[str]:
        """!
        @brief Get build commands.

        @returns list of commands.
        """
        return self.__build_commands

    def addStartCommand(self, cmd: str, fork: bool = False):
        """!
        @brief Add new command to start script.

        The command should not be a blocking command. If you need to run a
        blocking command, set fork to true and fork it to the background so
        that it won't block the execution of other commands.

        @param cmd command to add.
        @param fork (optional) fork to command to background?

        Use this to add start steps to the node. For example, if using the
        "docker" compiler, this will be added to start.sh.
        """
        self.__start_commands.append((cmd, fork))

    def getStartCommands(self) -> List[Tuple[str, bool]]:
        """!
        @brief Get start commands.

        @returns list of tuples, where the first element is command, and the
        second element indicates if this command should be forked.
        """
        return self.__start_commands

    def getInterfaces(self) -> List[Interface]:
        """!
        @brief Get list of interfaces.

        @returns list of interfaces.
        """
        return self.__interfaces
        
    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'Node {}:\n'.format(self.__name)

        indent += 4
        out += ' ' * indent
        out += 'Role: {}\n'.format(self.__role)
        out += ' ' * indent

        out += 'Interfaces:\n'
        for interface in self.__interfaces:
            out += interface.print(indent + 4)

        out += ' ' * indent
        out += 'Files:\n'
        for file in self.__files.values():
            out += file.print(indent + 4)

        out += ' ' * indent
        out += 'Software:\n'
        indent += 4
        for software in self.__softwares:
            out += ' ' * indent
            out += '{}\n'.format(software)
        indent -= 4

        out += ' ' * indent
        out += 'Additional Build Commands:\n'
        indent += 4
        for cmd in self.__build_commands:
            out += ' ' * indent
            out += '{}\n'.format(cmd)
        indent -= 4

        out += ' ' * indent
        out += 'Additional Start Commands:\n'
        indent += 4
        for (cmd, fork) in self.__start_commands:
            out += ' ' * indent
            out += '{}{}\n'.format(cmd, ' (fork)' if fork else '')
        indent -= 4

        return out



from os import mkdir, chdir, getcwd
from sys import stderr

class Compiler:
    """!
    @brief The Compiler base class.

    Compiler takes the rendered result and compiles them to working simulators.
    """

    def _doCompile(self):
        """!
        @brief Compiler driver implementation.

        This method should be implemented by the compiler driver. The driver
        class can assume that the current working directory is the output
        folder.

        @param registry rendered simulation.
        """
        raise NotImplementedError('_doCompile not implemented.')

    def getName(self) -> str:
        """!
        @brief Get the name of compiler driver.

        @returns name of the driver.
        """
        raise NotImplementedError('getName not implemented.')

    def compile(self, output: str):
        """!
        @brief Compile the simulation.

        @param output output directory path.
        """
        cur = getcwd()
        mkdir(output)
        chdir(output)
        self._doCompile()
        chdir(cur)

    def _log(self, message: str) -> None:
        """!
        @brief Log to stderr.
        """
        print("== {}Compiler: {}".format(self.getName(), message), file=stderr)



from .Layer import Layer
from seedsim.core import Network, Registry, Node
from seedsim.core.enums import NetworkType
from typing import Set, Dict, List

OspfFileTemplates: Dict[str, str] = {}

OspfFileTemplates['ospf_body'] = """
    ipv4 {{
        table t_ospf;
        import all;
        export all;
    }};
    area 0 {{
{interfaces}
    }};
"""

OspfFileTemplates['ospf_interface'] = """\
        interface "{interfaceName}" {{ hello 1; dead count 2; }};
"""

OspfFileTemplates['ospf_stub_interface'] = """\
        interface "{interfaceName}" {{ stub; }};
"""

class Ospf(Layer):
    """!
    @brief Ospf (OSPF) layer.

    @todo allow mask as

    This layer enables OSPF on all router nodes. By default, this will make all
    internal network interfaces (interfaces that are connected to a network
    created by BaseLayer::createNetwork) OSPF interface. Other interfaces like
    the IX interface will also be added as stub interface.
    """

    __stubs: Set[Network]
    __masked: Set[Network]
    __masked_asn: Set[int]
    __reg = Registry()

    def __init__(self):
        """!
        @brief Ospf (OSPF) layer conscrutor.
        """

        self.__stubs = set()
        self.__masked = set()
        self.__masked_asn = set()

        self.addDependency('Routing', False, False)

    def getName(self) -> str:
        return 'Ospf'

    def markStub(self, net: Network):
        """!
        @brief Set all OSPF interfaces connected to a network as stub
        interfaces.

        By default, all internal networks will be active OSPF interface. This
        method can be used to override the behavior and make the interface
        stub interface (i.e., passive). For example, you can mark host-only 
        internal networks as a stub.

        @param net Network.
        @throws AssertionError if network is not local
        """
        assert net.getType() != NetworkType.InternetExchange, 'cannot operator on IX network.'
        self.__stubs.add(net)

    def markStubByName(self, asn: int, netname: str):
        """!
        @brief Set all OSPF interfaces connected to a network as stub
        interfaces.

        By default, all internal networks will be active OSPF interface. This
        method can be used to override the behavior and make the interface
        stub interface (i.e., passive). For example, you can mark host-only 
        internal networks as a stub.

        @param asn ASN to operate on.
        @param netname name of the network.
        """
        self.markStub(self.__reg.get(str(asn), 'net', netname))

    def maskNetwork(self, net: Network):
        """!
        @brief Remove all OSPF interfaces connected to a network.

        By default, all internal networks will be active OSPF interface. Use
        this method to mask a network and disable OSPF on all connected
        interface.

        @todo handle IX LAN masking?

        @param net network.
        @throws AssertionError if network is not local.
        """
        assert net.getType() != NetworkType.InternetExchange, 'cannot mask IX network.'
        self.__masked.add(net)

    def maskAsn(self, asn: int):
        """!
        @brief Disable OSPF for an AS.

        @param asn asn.
        """
        self.__masked_asn.add(asn)

    def maskByName(self, asn: int, netname: str):
        """!
        @brief Remove all OSPF interfaces connected to a network.

        By default, all internal networks will be active OSPF interface. Use
        this method to mask a network and disable OSPF on all connected
        interface.

        @param asn ASN to operate on.
        @param netname name of the network.
        """
        self.maskNetwork(self.__reg.get(str(asn), 'net', netname))

    def isMasked(self, net: Network) -> bool:
        """!
        @brief Test if a network is masked.

        @param net network to test.
        
        @returns if net is masked.
        """
        return net in self.__masked

    def onRender(self):
        for ((scope, type, name), obj) in self.__reg.getAll().items():
            if type != 'rnode': continue
            router: Node = obj
            if router.getAsn() in self.__masked_asn: continue

            stubs: List[str] = ['dummy0']
            active: List[str] = []

            self._log('setting up OSPF for router as{}/{}...'.format(scope, name))
            for iface in router.getInterfaces():
                net = iface.getNet()

                if net in self.__masked: continue

                if net in self.__stubs or net.getType() == NetworkType.InternetExchange:
                    stubs.append(net.getName())
                    continue

                active.append(net.getName())
            
            ospf_interfaces = ''
            for name in stubs: ospf_interfaces += OspfFileTemplates['ospf_stub_interface'].format(
                interfaceName = name
            )
            for name in active: ospf_interfaces += OspfFileTemplates['ospf_interface'].format(
                interfaceName = name
            )

            if ospf_interfaces != '':
                router.addTable('t_ospf')
                router.addProtocol('ospf', 'ospf1', OspfFileTemplates['ospf_body'].format(
                    interfaces = ospf_interfaces
                ))
                router.addTablePipe('t_ospf')

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'OspfLayer:\n'

        indent += 4

        out += ' ' * indent
        out += 'Stub Networks:\n'
        indent += 4
        for net in self.__stubs:
            out += ' ' * indent
            (scope, _, netname) = net.getRegistryInfo()
            out += 'as{}/{}\n'.format(scope, netname)
        indent -= 4

        out += ' ' * indent
        out += 'Masked Networks:\n'
        indent += 4
        for net in self.__masked:
            out += ' ' * indent
            (scope, _, netname) = net.getRegistryInfo()
            out += 'as{}/{}\n'.format(scope, netname)
        indent -= 4

        out += ' ' * indent
        out += 'Masked AS:\n'
        indent += 4
        for asn in self.__masked_asn:
            out += ' ' * indent
            out += 'as{}\n'.format(asn)

        return out




from __future__ import annotations
from .Service import Service, Server
from seedsim.core import Node, Printable, ScopedRegistry
from seedsim.core.enums import NodeRole, NetworkType
from typing import List, Dict, Tuple, Set
from re import sub
from random import randint

DomainNameServiceFileTemplates: Dict[str, str] = {}

DomainNameServiceFileTemplates['named_options'] = '''\
options {
	directory "/var/cache/bind";
	recursion no;
	dnssec-validation no;
    empty-zones-enable no;
	allow-query { any; };
    allow-update { any; };
};
'''

class Zone(Printable):
    """!
    @brief Domain name zone.
    """
    __zonename: str
    __subzones: Dict[str, Zone]
    __records: List[str]
    __gules: List[str]

    def __init__(self, name: str):
        """!
        @brief Zone constructor.
        
        @param name full zonename.
        """
        self.__zonename = name
        self.__subzones = {}
        self.__records = [
            '$TTL 300',
            '$ORIGIN {}'.format(name if name != '' else '.')
        ]
        self.__gules = []

    def getName(self) -> str:
        """!
        @brief Get zonename.

        @returns zonename.
        """
        return self.__zonename

    def getSubZone(self, name: str) -> Zone:
        """!
        @brief Get a subzone, if not exists, a new one will be created.

        @param name partial zonename. For example, if current zone is "com.", to
        get "example.com.", use getSubZone("example")

        @returns zone.
        @throws AssertionError if invalid zonename.
        """
        assert '.' not in name, 'invalid subzone name "{}"'.format(name)
        if name in self.__subzones: return self.__subzones[name]
        self.__subzones[name] = Zone('{}.{}'.format(name, self.__zonename))
        return self.__subzones[name]
    
    def getSubZones(self) -> Dict[str, Zone]:
        """!
        @brief Get all subzones.

        @return subzones dict.
        """
        return self.__subzones

    def addRecord(self, record: str):
        """!
        @brief Add a new record to zone.

        @todo NS?
        """
        self.__records.append(record)

    def addGuleRecord(self, fqdn: str, addr: str):
        """!
        @brief Add a new gule record.

        Use this method to register a name server in the parent zone.

        @param fqdn full domain name of the name server.
        @param addr IP address of the name server.
        """
        if fqdn[-1] != '.': fqdn += '.'
        zonename = self.__zonename if self.__zonename != '' else '.' 
        self.__gules.append('{} A {}'.format(fqdn, addr))
        self.__gules.append('{} NS {}'.format(zonename, fqdn))

    def resolveTo(self, name: str, node: Node):
        """!
        @brief Add a new A record, pointing to the given node.

        @param name name.
        @param node node.

        @throws AssertionError if node does not have valid interfaces.
        """

        address: str = None
        ifaces = node.getInterfaces()
        assert len(ifaces) > 0, 'Node has no interfaces.'
        for iface in ifaces:
            net = iface.getNet()
            if net.getType() == NetworkType.Host or net.getType() == NetworkType.Local:
                address = iface.getAddress()
                break

        assert address != None, 'Node has no valid interfaces.'
        self.__records.append('{} A {}'.format(name, address))

    def getRecords(self) -> List[str]:
        """!
        @brief Get all records.

        @return list of records.
        """
        return self.__records

    def getGuleRecords(self) -> List[str]:
        """!
        @brief Get all gule records.

        @return list of records.
        """
        return self.__gules

    def findRecords(self, keyword: str) -> List[str]:
        """!
        @brief Find a record.

        @param keyword keyword.

        @return list of records.
        """
        return [ r for r in self.__records if keyword in r ]

    def print(self, indent: int) -> str:
        out = ' ' * indent
        zonename = self.__zonename if self.__zonename != '' else '(root zone)'
        out += 'Zone "{}":\n'.format(zonename)

        indent += 4
        out += ' ' * indent
        out += 'Zonefile:\n'

        indent += 4
        for record in self.__records:
            out += ' ' * indent
            out += '{}\n'.format(record)

        indent -= 4
        out += ' ' * indent
        out += 'Subzones:\n'
        
        indent += 4
        for subzone in self.__subzones.values():
            out += subzone.print(indent)

        return out

class DomainNameServer(Server):
    """!
    @brief The domain name server.
    """

    __node: Node
    __zones: Set[Zone]

    def __init__(self, node: Node):
        """!
        @brief DomainNameServer constructor.

        @param node node to install server on.
        """
        self.__node = node
        self.__zones = set()

    def addZone(self, zone: Zone):
        """!
        @brief Add a zone to this node.

        You should use DomainNameService.hostZoneOn to host zone on node if you
        want the automated NS record to work.
        """
        self.__zones.add(zone)

    def getZones(self) -> List[Zone]:
        """!
        @brief Get list of zones hosted on the node.

        @returns list of zones.
        """
        return list(self.__zones)

    def getNode(self) -> Node:
        """!
        @brief Get the node.

        @returns node.
        """
        return self.__node

    def print(self, indent: int) -> str:
        out = ' ' * indent
        (scope, _, name) = self.__node.getRegistryInfo()
        out += 'Zones on as{}/{}:\n'.format(scope, name)
        indent += 4
        for zone in self.__zones:
            out += ' ' * indent
            zname = zone.getName()
            if zname == '' or zname[-1] != '.': zname += '.'
            out += '{}\n'.format(zname)

        return out

    def install(self):
        """!
        @brief Handle the installation.
        """
        self.__node.addSoftware('bind9')
        self.__node.addStartCommand('echo "include \\"/etc/bind/named.conf.zones\\";" >> /etc/bind/named.conf.local')
        self.__node.setFile('/etc/bind/named.conf.options', DomainNameServiceFileTemplates['named_options'])

        for zone in self.__zones:
            zonename = filename = zone.getName()
            if zonename == '' or zonename == '.':
                filename = 'root'
                zonename = '.'
            zonepath = '/etc/bind/zones/{}'.format(filename)
            self.__node.setFile(zonepath, '\n'.join(zone.getRecords()))
            self.__node.appendFile('/etc/bind/named.conf.zones',
                'zone "{}" {{ type master; file "{}"; allow-update {{ any; }}; }};\n'.format(zonename, zonepath)
            )

        self.__node.addStartCommand('chown -R bind:bind /etc/bind/zones')
        self.__node.addStartCommand('service named start')
    
class DomainNameService(Service):
    """!
    @brief The domain name service.
    """

    __rootZone = Zone('') # singleton
    __servers: List[DomainNameServer]
    __reg = ScopedRegistry('seedsim')
    __autoNs: bool

    def __init__(self, autoNameServer: bool = True):
        """!
        @brief DomainNameService constructor.
        
        @param autoNameServer add gule records to parents automaically.
        """
        self.__autoNs = autoNameServer
        self.__servers = []
        self.addDependency('Base', False, False)

    def getName(self):
        return 'DomainNameService'

    def getConflicts(self) -> List[str]:
        return ['DomainNameCachingService']
    
    def getZone(self, domain: str) -> Zone:
        """!
        @brief Get a zone, create it if not exist.

        This method only create the zone. Host it with hostZoneOn.

        @param domain zone name.

        @returns zone handler.
        """
        if domain == '.' or domain == '': return self.__rootZone
        path: List[str] = sub(r'\.$', '', domain).split('.')
        path.reverse()
        zoneptr = self.__rootZone
        for z in path:
            zoneptr = zoneptr.getSubZone(z)

        return zoneptr

    def getRootZone(self) -> Zone:
        """!
        @brief Get the root zone.

        @return root zone.
        """
        return self.__rootZone

    def _doInstall(self, node: Node) -> DomainNameServer:
        server: DomainNameServer = node.getAttribute('__domain_name_service_server')
        if server != None: return server
        server = DomainNameServer(node)
        self.__servers.append(server)
        node.setAttribute('__domain_name_service_server', server)
        return server

    def hostZoneOn(self, domain: str, node: Node, addNsAndSoa: bool = True):
        """!
        @brief Host a zone on the given node.

        Zone must be created with getZone first.

        @param domain zone name.
        @param node target node.
        @param addNsAndSoa (optional) add NS (ns1.domain.tld) and SOA records
        automatically, true by default. This method will also add gule records.

        @throws AssertionError if node is not a host node.
        @throws AssertionError if node has no interface.
        """
        if not node.hasAttribute('__domain_name_service_server'):
            self.installOn(node)

        server: DomainNameServer = node.getAttribute('__domain_name_service_server')
        zone = self.getZone(domain)
        server.addZone(zone)

        if not addNsAndSoa: return

        ifaces = node.getInterfaces()
        assert len(ifaces) > 0, 'node has not interfaces'
        addr = ifaces[0].getAddress()

        if domain[-1] != '.': domain += '.'
        if domain == '.': domain = ''

        if len(zone.findRecords('SOA')) == 0:
            zone.addRecord('@ SOA {} {} {} 900 900 1800 60'.format('ns1.{}'.format(domain), 'admin.{}'.format(domain), randint(1, 0xffffffff)))

        zone.addGuleRecord('ns1.{}'.format(domain), addr)
        zone.addRecord('ns1.{} A {}'.format(domain, addr))
        zone.addRecord('@ NS ns1.{}'.format(domain))

    def getServerByZoneName(self, zonename: str) -> DomainNameServer:
        """!
        @brief Get server by zonename.

        @param zonename zonename.

        @returns server if exists, None otherwise.
        """

        if zonename[-1] != '.': zonename += '.'
        if zonename == '.': zonename = ''
        for server in self.__servers:
            for zone in server.getZones():
                if zone.getName() == zonename: return server

        return None

    def __autoNameServer(self, zone: Zone):
        """!
        @brief Try to automatically add NS records of children to parent zones.

        @param zone root zone reference.
        """
        if (len(zone.getSubZones().values()) == 0): return
        self._log('Collecting subzones NSes of "{}"...'.format(zone.getName()))
        for subzone in zone.getSubZones().values():
            for gule in subzone.getGuleRecords(): zone.addRecord(gule)
            self.__autoNameServer(subzone)

    def onRender(self):
        if self.__autoNs:
            self._log('Setting up NS records...')
            self.__autoNameServer(self.__rootZone)

        for server in self.__servers:
            server.install()

    def print(self, indent: int) -> str:
        out = ' ' * indent
        out += 'DomainNameService:\n'

        indent += 4
        out += self.__rootZone.print(indent)

        out += ' ' * indent
        out += 'Hosted zones:\n'

        for server in self.__servers:
            out += server.print(indent + 4)

        return out